import copy

from typing import Dict, Union

import logging

from multiqc.plots import linegraph, table

log = logging.getLogger(__name__)


class CoverageReportMixin:
    def parse_samtools_coverage(self):
        """Find Samtools coverage logs and parse their data"""

        data_by_sample = dict()
        for f in self.find_log_files("samtools/coverage"):
            metrics_by_chrom = parse_single_report(f)
            if len(metrics_by_chrom) > 0:
                if f["s_name"] in data_by_sample:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="coverage")
                data_by_sample[f["s_name"]] = metrics_by_chrom

        # Filter to strip out ignored sample names
        data_by_sample = self.ignore_samples(data_by_sample)
        if len(data_by_sample) == 0:
            return 0

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file (restructure first)
        self.write_data_file(data_by_sample, "multiqc_samtools_coverage")

        # Make a table/violin summarising the stats across all regions (mean or sum)
        self.summary_table(data_by_sample)

        # Make a line plot showing coverage stats per region, with a tab switch between stats
        self.lineplot_per_region(data_by_sample)

        # Return the number of logs that were found
        return len(data_by_sample)

    def summary_table(self, data_by_sample):
        table_data = {sname: {} for sname in data_by_sample}
        for sample, d_by_chrom in data_by_sample.items():
            vals = list(d_by_chrom.values())
            table_data[sample]["numreads"] = sum([m["numreads"] for m in vals])
            table_data[sample]["covbases"] = sum([m["covbases"] for m in vals])
            for m in vals:
                m["size"] = m["endpos"] - m["startpos"] + 1
            total_size = sum([m["size"] for m in vals])
            # Average weighted by size. Multiplying by individual weight and dividing by total weight
            table_data[sample]["coverage"] = sum([m["coverage"] * m["size"] for m in vals]) / total_size
            table_data[sample]["meandepth"] = sum([m["meandepth"] * m["size"] for m in vals]) / total_size
            table_data[sample]["meanbaseq"] = sum([m["meanbaseq"] * m["size"] for m in vals]) / total_size
            table_data[sample]["meanmapq"] = sum([m["meanmapq"] * m["size"] for m in vals]) / total_size

        headers = {
            "numreads": {
                "title": "Reads",
                "description": "Total number of mapped reads",
                "shared_key": "read_count",
                "scale": "RdYlGn",
            },
            "covbases": {
                "title": "Bases",
                "description": "Total number of mapped base pairs",
                "shared_key": "base_count",
                "scale": "Blues",
            },
            "coverage": {
                "title": "Coverage",
                "description": "Percentage of region covered with reads",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "YlGn",
            },
            "meandepth": {
                "title": "Mean depth",
                "description": "Mean depth of coverage",
                "min": 0,
                "suffix": "x",
                "scale": "RdYlGn",
            },
            "meanbaseq": {
                "title": "Mean BQ",
                "description": "Mean base quality",
                "min": 0,
                "scale": "Blues",
            },
            "meanmapq": {
                "title": "Mean MQ",
                "description": "Mean mapping quality",
                "min": 0,
                "max": 60,
                "scale": "RdYlGn",
            },
        }

        self.add_section(
            name="Coverage: global stats",
            anchor="samtools-coverage-table-section",
            description=(
                "Stats parsed from <code>samtools coverage</code> output, and summarized "
                "(added up or weighted-averaged) across all regions."
            ),
            plot=table.plot(
                table_data,
                copy.deepcopy(headers),
                pconfig={
                    "id": "samtools-coverage-table",
                    "title": "Samtools Coverage: Summary",
                },
            ),
        )

        for h in headers:
            headers[h]["hidden"] = True
        headers["meandepth"]["hidden"] = False
        self.general_stats_addcols(table_data, headers, namespace="coverage")

    def lineplot_per_region(self, data_by_sample):
        tabs = {
            "numreads": {
                "title": "Mapped reads per region",
                "name": "Reads",
                "ylab": "# reads",
                "tt_suffix": " reads",
                "tt_decimals": 0,
            },
            "covbases": {
                "title": "Mapped base pairs per region",
                "name": "Bases",
                "ylab": "# bases",
                "tt_suffix": " bp",
                "tt_decimals": 0,
            },
            "coverage": {
                "title": "Percentage of region covered with reads",
                "name": "Coverage",
                "ylab": "Coverage %",
                "tt_suffix": "%",
            },
            "meandepth": {
                "title": "Mean depth per region",
                "name": "Mean depth",
                "ylab": "Depth",
                "tt_suffix": "x",
            },
            "meanbaseq": {
                "title": "Mean base quality per region",
                "name": "BQ",
                "ylab": "Base quality",
                "tt_suffix": "",
            },
            "meanmapq": {
                "title": "Mean mapping quality per region",
                "name": "MQ",
                "ylab": "Mapping quality",
                "tt_suffix": "",
            },
        }
        datasets = [
            {
                sample: {chrom: metrics[metric] for chrom, metrics in data_by_sample[sample].items()}
                for sample in data_by_sample
            }
            for metric in tabs
        ]
        data_labels = list(tabs.values())
        for dconfig in data_labels:
            dconfig["title"] = "Samtools coverage: " + dconfig["title"]

        self.add_section(
            name="Coverage: stats per region",
            description="Per-region stats parsed from <code>samtools coverage</code> output.",
            anchor="samtools-coverage-section",
            plot=linegraph.plot(
                datasets,
                {
                    "id": "samtools-coverage",
                    "title": data_labels[0]["title"],
                    "xlab": "Region",
                    "ylab": data_labels[0]["ylab"],
                    "ymin": 0,
                    "categories": True,
                    "smooth_points": 500,
                    "logswitch": True,
                    "hide_zero_cats": False,
                    "data_labels": data_labels,
                },
            ),
        )


EXPECTED_COLUMNS = [
    "rname",
    "startpos",
    "endpos",
    "numreads",
    "covbases",
    "coverage",
    "meandepth",
    "meanbaseq",
    "meanmapq",
]


def parse_single_report(f) -> Dict[str, Dict[str, Union[int, float]]]:
    """
    Example:
    #rname	startpos	endpos	numreads	covbases	coverage	meandepth	meanbaseq	meanmapq
    oligo_1512_adapters	1	156	10	156	100	7.26282	18.4	29.6
    oligo_741_adapters	1	156	0	0	0	0	0	0

    Returns a dictionary with the contig name (rname) as the key and the rest of the fields as a dictionary
    """
    parsed_data = {}
    lines = f["f"].splitlines()
    expected_header = "#" + "\t".join(EXPECTED_COLUMNS)
    if lines[0] != expected_header:
        logging.warning(f"Expected header for samtools coverage: {expected_header}, got: {lines[0]}")
        return {}

    for line in lines[1:]:
        fields = line.strip().split("\t")
        if len(fields) != len(EXPECTED_COLUMNS):
            logging.warning(f"Skipping line with {len(fields)} fields, expected {len(EXPECTED_COLUMNS)}: {line}")
        rname, startpos, endpos, numreads, covbases, coverage, meandepth, meanbaseq, meanmapq = fields
        parsed_data[rname] = dict(
            startpos=int(startpos),
            endpos=int(endpos),
            numreads=int(numreads),
            covbases=int(covbases),
            coverage=float(coverage),
            meandepth=float(meandepth),
            meanbaseq=float(meanbaseq),
            meanmapq=float(meanmapq),
        )

    return parsed_data



import logging

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import table

from .utils import Metric, exist_and_number, make_headers

# Initialise the logger
log = logging.getLogger(__name__)


NAMESPACE = "Variant calling"


class DragenVCMetrics(BaseMultiqcModule):
    def add_vc_metrics(self):
        vc_data_by_sample = dict()
        for f in self.find_log_files("dragen/vc_metrics"):
            data = parse_vc_metrics_file(f)
            s_name = f["s_name"]
            if s_name in vc_data_by_sample:
                log.debug(f"Duplicate DRAGEN vc_metrics file was found! Overwriting sample {s_name}")
            self.add_data_source(f, s_name=s_name, section="vc_metrics")
            vc_data_by_sample[s_name] = data

        gvcf_data_by_sample = dict()
        for f in self.find_log_files("dragen/gvcf_metrics"):
            data = parse_vc_metrics_file(f)
            s_name = f["s_name"]
            if s_name in gvcf_data_by_sample:
                log.debug(f"Duplicate DRAGEN gvcf_metrics file was found! Overwriting sample {s_name}")
            self.add_data_source(f, s_name=s_name, section="gvcf_metrics")
            gvcf_data_by_sample[s_name] = data

        # Filter to strip out ignored sample names:
        vc_data_by_sample = self.ignore_samples(vc_data_by_sample)
        gvcf_data_by_sample = self.ignore_samples(gvcf_data_by_sample)
        if not vc_data_by_sample:
            return set()

        # Write data to file
        self.write_data_file(vc_data_by_sample, "dragen_vc_metrics")
        if gvcf_data_by_sample:
            self.write_data_file(gvcf_data_by_sample, "dragen_gvcf_metrics")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        all_metric_names = set()
        for sn, sdata in vc_data_by_sample.items():
            for m in sdata.keys():
                all_metric_names.add(m)

        gen_stats_headers, table_headers = make_headers(all_metric_names, VC_METRICS)

        self.general_stats_addcols(vc_data_by_sample, gen_stats_headers, namespace=NAMESPACE)

        self.add_section(
            name="Variant calling",
            anchor="dragen-vc-metrics",
            description="""
            Metrics are reported for each sample in multi-sample VCF
            files. Based on the run case, metrics are reported either as standard
            VARIANT CALLER or JOINT CALLER. All metrics are reported for post-filter VCFs,
            except for the "Filtered" metrics which represent how many variants were filtered out
            from pre-filter VCF to generate the post-filter VCF.
            """,
            plot=table.plot(
                vc_data_by_sample,
                table_headers,
                pconfig={
                    "id": "dragen-vc-metrics-table",
                    "namespace": NAMESPACE,
                    "title": "DRAGEN: Variant calling metrics",
                },
            ),
        )

        if gvcf_data_by_sample:
            self.add_section(
                name="GVCF metrics",
                anchor="dragen-gvcf-metrics",
                description="""
                Metrics are calculated for each sample in a corresponding GVCF file.
                """,
                plot=table.plot(
                    gvcf_data_by_sample,
                    table_headers,
                    pconfig={
                        "id": "dragen-gvcf-metrics-table",
                        "namespace": NAMESPACE,
                        "title": "DRAGEN: GVCF metrics",
                    },
                ),
            )

        return vc_data_by_sample.keys()


VC_METRICS = [
    Metric(
        m.id,
        m.title,
        in_genstats=m.in_genstats,
        in_own_tabl=m.in_own_tabl,
        descr=m.descr,
        unit=m.unit,
        namespace=m.namespace or NAMESPACE,
        the_higher_the_worse=m.the_higher_the_worse,
    )
    for m in [
        # id_in_data                                        title (display name)   gen_stats  vc_table  unit description
        # Read stats:
        Metric("Total", "Variants", "#", "#", "", "Total number of variants (SNPs + MNPs + INDELS)."),
        Metric(
            "Biallelic",
            "Biallelic",
            None,
            "hid",
            "",
            "Number of sites in a genome that contains two observed alleles, counting the reference as one, and therefore allowing for one variant allele",
        ),
        Metric(
            "Multiallelic",
            "Multiallelic",
            "hid",
            "%",
            "",
            "Number of sites in the VCF that contain three or more observed alleles. The reference is counted as one, therefore allowing for two or more variant alleles",
        ),
        Metric(
            "SNPs",
            "SNP",
            "hid",
            "%",
            "",
            "Number of SNPs in the variant set. A variant is counted as an SNP when the reference, allele 1, and allele2 are all length 1",
        ),
        Metric("Indels", "Indel", "hid", "hid", "", "Number of insetions and deletions in the variant set."),
        Metric("Insertions", "Ins", None, "%", "", "Number of insetions in the variant set."),
        Metric("Deletions", "Del", None, "%", "", "Number of deletions in the variant set."),
        Metric(
            "Insertions (Hom)", "Hom ins", None, "hid", "", "Number of variants that contains homozygous insertions"
        ),
        Metric(
            "Insertions (Het)",
            "Het ins",
            None,
            "hid",
            "",
            "Number of variants where both alleles are insertions, but not homozygous",
        ),
        Metric("Deletions (Hom)", "Hom del", None, "hid", "", "Number of variants that contains homozygous deletions"),
        Metric(
            "Deletions (Het)",
            "Het del",
            None,
            "hid",
            "",
            "Number of variants where both alleles are deletion, but not homozygous",
        ),
        Metric(
            "Indels (Het)",
            "Het indel",
            None,
            "hid",
            "",
            "Number of variants where genotypes are either [insertion+deletion], [insertion+snp] or [deletion+snp].",
        ),
        Metric(
            "DeNovo SNPs",
            "DeNovo SNPs",
            None,
            None,
            "",
            "Number of DeNovo marked SNPs, with DQ > 0.05. Set the --qc-snp-denovo-quality-threshold option to the required threshold. The default is 0.05.",
        ),
        Metric(
            "DeNovo INDELs",
            "DeNovo indel",
            None,
            None,
            "",
            "Number of DeNovo marked indels, with DQ > 0.05. Set the --qc-snp-denovo-quality-threshold option to the required threshold. The default is 0.05.",
        ),
        Metric(
            "DeNovo MNPs",
            "DeNovo MNPs",
            None,
            None,
            "",
            "Number of DeNovo marked MNPs, with DQ > 0.05. Set the --qc-snp-denovo-quality-threshold option to the required threshold. The default is 0.05.",
        ),
        Metric(
            "Chr X number of SNPs over genome",
            "ChrX SNP",
            None,
            None,
            "",
            "Number of SNPs in chromosome X (or in the intersection of chromosome X with the target region). " "",
            "If there was no alignment to either chromosome X, this metric shows as NA",
        ),
        Metric(
            "Chr Y number of SNPs over genome",
            "ChrY SNP",
            None,
            None,
            "",
            "Number of SNPs in chromosome Y (or in the intersection of chromosome Y with the target region). " "",
            "If there was no alignment to either chromosome Y, this metric shows as NA",
        ),
        Metric(
            "(Chr X SNPs)/(chr Y SNPs) ratio over genome",
            "X/Y SNP ratio",
            None,
            "hid",
            "",
            "Number of SNPs in chromosome X (or in the intersection of chromosome X with the target region) " "",
            "divided by the number of SNPs in chromosome Y (or in the intersection of chromosome Y with the " "",
            "target region). If there was no alignment to either chromosome X or chromosome Y, this metric " "",
            "shows as NA",
        ),
        Metric(
            "SNP Transitions",
            "SNP Ti",
            None,
            None,
            "",
            "Number of transitions - interchanges of two purines (A<->G) or two pyrimidines (C<->T)",
        ),
        Metric(
            "SNP Transversions",
            "SNP Tv",
            None,
            None,
            "",
            "Number of transversions - interchanges of purine and pyrimidine bases",
        ),
        Metric("Ti/Tv ratio", "Ti/Tv", "hid", "#", "", "Ti/Tv ratio: ratio of transitions to transitions."),
        Metric("Heterozygous", "Het", "hid", "hid", "", "Number of heterozygous variants"),
        Metric("Homozygous", "Hom", "hid", "hid", "", "Number of homozygous variants"),
        Metric("Het/Hom ratio", "Het/Hom", "hid", "#", "", "Heterozygous/ homozygous ratio"),
        Metric(
            "In dbSNP",
            "In dbSNP",
            None,
            None,
            "",
            "Number of variants detected that are present in the dbsnp reference file. If no dbsnp file " "",
            "is provided via the --bsnp option, then both the In dbSNP and Novel metrics show as NA.",
            the_higher_the_worse=True,
        ),
        Metric(
            "Not in dbSNP",
            "Novel",
            None,
            None,
            "",
            "Number of all variants minus number of variants in dbSNP. If no dbsnp file " "",
            "is provided via the --bsnp option, then both the In dbSNP and Novel metrics show as NA.",
        ),
        Metric(
            "Percent Callability",
            "Callability",
            None,
            "#",
            "",
            "Available only in germline mode with gVCF output. The percentage of non-N reference " "",
            "positions having a PASSing genotype call. Multi-allelic variants are not counted. " "",
            "Deletions are counted for all the deleted reference positions only for homozygous calls. " "",
            "Only autosomes and chromosomes X, Y and M are considered.",
        ),
        Metric(
            "Percent Autosome Callability",
            "Autosome callability",
            None,
            "hid",
            "",
            "Available only in germline mode with gVCF output. The percentage of non-N reference " "",
            "positions having a PASSing genotype call. Multi-allelic variants are not counted. " "",
            "Deletions are counted for all the deleted reference positions only for homozygous calls. " "",
            "Only autosomes are considered (for all chromosomes, see the Callability metric).",
        ),
        Metric(
            "Filtered vars",
            "Filt var",
            "hid",
            "hid",
            "",
            "Number of raw variants minus the number of PASSed variants",
            the_higher_the_worse=True,
        ),
        Metric(
            "Filtered SNPs",
            "Filt SNP",
            "hid",
            "%",
            "",
            "Number of raw SNPs minus the number of PASSed SNPs",
            the_higher_the_worse=True,
        ),
        Metric(
            "Filtered indels",
            "Filt indel",
            "hid",
            "%",
            "",
            "Number of raw indels minus the number of PASSed indels",
            the_higher_the_worse=True,
        ),
        Metric(
            "Reads Processed",
            "VC reads",
            None,
            "#",
            "reads",
            "The number of reads used for variant calling, excluding any duplicate marked reads and reads falling outside of the target region",
        ),
    ]
]


def parse_vc_metrics_file(f):
    """
    T_SRR7890936_50pc.vc_metrics.csv or T_SRR7890936_50pc.gvcf_metrics.csv

    VARIANT CALLER SUMMARY,,Number of samples,1
    VARIANT CALLER SUMMARY,,Reads Processed,2721782043
    VARIANT CALLER SUMMARY,,Child Sample,NA
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Total,170013,100.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Biallelic,170013,100.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Multiallelic,0,0.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,SNPs,138978,81.75
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Insertions (Hom),0,0.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Insertions (Het),14291,8.41
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Deletions (Hom),0,0.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Deletions (Het),16744,9.85
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Indels (Het),0,0.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Chr X number of SNPs over genome,32487
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Chr Y number of SNPs over genome,131
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,(Chr X SNPs)/(chr Y SNPs) ratio over genome,247.99
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,SNP Transitions,79948
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,SNP Transversions,59025
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Ti/Tv ratio,1.35
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Heterozygous,170013
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Homozygous,0
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Het/Hom ratio,0.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,In dbSNP,0,0.00
    VARIANT CALLER PREFILTER,T_SRR7890936_50pc,Not in dbSNP,0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Total,123219,100.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Biallelic,123219,100.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Multiallelic,0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,SNPs,104900,85.13
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Insertions (Hom),0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Insertions (Het),8060,6.54
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Deletions (Hom),0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Deletions (Het),10259,8.33
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Indels (Het),0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Chr X number of SNPs over genome,28162
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Chr Y number of SNPs over genome,8
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,(Chr X SNPs)/(chr Y SNPs) ratio over genome,3520.25
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,SNP Transitions,62111
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,SNP Transversions,42789
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Ti/Tv ratio,1.45
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Heterozygous,123219
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Homozygous,0
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Het/Hom ratio,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,In dbSNP,0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Not in dbSNP,0,0.00
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Percent Callability,NA
    VARIANT CALLER POSTFILTER,T_SRR7890936_50pc,Percent Autosome Callability,NA
    """

    summary_data = dict()
    prefilter_data = dict()
    postfilter_data = dict()

    for line in f["f"].splitlines():
        fields = line.split(",")
        analysis = fields[0]
        # sample = fields[1]
        metric = fields[2]
        value = fields[3]

        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                pass

        percentage = None
        if len(fields) > 4:  # percentage
            percentage = fields[4]
            try:
                percentage = float(percentage)
            except ValueError:
                pass

        if analysis == "VARIANT CALLER SUMMARY":
            summary_data[metric] = value

        if analysis == "VARIANT CALLER PREFILTER":
            prefilter_data[metric] = value

        if analysis in ["VARIANT CALLER POSTFILTER", "VARIANT CALLER POSTFILTER GVCF"]:
            postfilter_data[metric] = value
            if percentage is not None:
                postfilter_data[metric + " pct"] = percentage

    # adding few more metrics: total insertions, deletions and indels numbers
    for data in [prefilter_data, postfilter_data]:
        if exist_and_number(data, "Insertions (Hom)", "Insertions (Het)"):
            data["Insertions"] = data["Insertions (Hom)"] + data["Insertions (Het)"]

        if exist_and_number(data, "Deletions (Hom)", "Deletions (Het)"):
            data["Deletions"] = data["Deletions (Hom)"] + data["Deletions (Het)"]

        if exist_and_number(data, "Insertions", "Deletions"):
            data["Indels"] = data["Insertions"] + data["Deletions"]

        if exist_and_number(data, "Total") and data["Total"] != 0:
            if exist_and_number(data, "Insertions"):
                data["Insertions pct"] = data["Insertions"] / data["Total"] * 100.0
            if exist_and_number(data, "Deletions"):
                data["Deletions pct"] = data["Deletions"] / data["Total"] * 100.0
            if exist_and_number(data, "Indels"):
                data["Indels pct"] = data["Indels"] / data["Total"] * 100.0

    data = postfilter_data
    data.update(summary_data)

    # we are not really interested in all the details of pre-filtered variants, however
    # it would be nice to report how much we filtered out
    if exist_and_number(data, "Total") and exist_and_number(prefilter_data, "Total"):
        data["Filtered vars"] = prefilter_data["Total"] - data["Total"]

    if exist_and_number(data, "SNPs") and exist_and_number(prefilter_data, "SNPs"):
        data["Filtered SNPs"] = prefilter_data["SNPs"] - data["SNPs"]

    if exist_and_number(data, "Indels") and exist_and_number(prefilter_data, "Indels"):
        data["Filtered indels"] = prefilter_data["Indels"] - data["Indels"]

    if (
        exist_and_number(prefilter_data, "Total")
        and exist_and_number(data, "Filtered vars")
        and prefilter_data["Total"] != 0
    ):
        data["Filtered vars pct"] = data["Filtered vars"] / prefilter_data["Total"] * 100.0

    if (
        exist_and_number(prefilter_data, "SNPs")
        and exist_and_number(data, "Filtered SNPs")
        and prefilter_data["SNPs"] != 0
    ):
        data["Filtered SNPs pct"] = data["Filtered SNPs"] / prefilter_data["SNPs"] * 100.0

    if (
        exist_and_number(prefilter_data, "Indels")
        and exist_and_number(data, "Filtered Indels")
        and prefilter_data["Indels"] != 0
    ):
        data["Filtered indels pct"] = data["Filtered indels"] / prefilter_data["Indels"] * 100.0

    return data



""" MultiQC module to parse output from HOPS postprocessing script """


import json
import logging
from json import JSONDecodeError

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import heatmap

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="HOPS",
            anchor="hops",
            href="https://github.com/rhuebler/HOPS/",
            info="is an ancient DNA characteristics screening tool of output from the metagenomic aligner MALT.",
            doi="10.1186/s13059-019-1903-0",
        )

        # Find and load any HOPS post-processing JSONs
        self.hops_data = dict()

        for f in self.find_log_files("hops", filehandles=True):
            try:
                self.parseJSON(f)
            except KeyError:
                logging.warning(f"Error loading file {f['fn']}")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        self.hops_data = self.ignore_samples(self.hops_data)

        if len(self.hops_data) == 0:
            raise ModuleNoSamplesFound

        # Write data to file
        self.write_data_file(self.hops_data, "hops")

        log.info(f"Found {len(self.hops_data)} samples")

        # This type of data isn't 'summarise-able' for general stats, so
        # skipping straight to heatmap. We also won't write data file to the
        # multiqc_data directory because it would be exactly same as input JSON.
        self.hops_heatmap()

    def parseJSON(self, f):
        """Parse the JSON output from HOPS and save the summary statistics"""

        try:
            parsed_json = json.load(f["f"])
        except JSONDecodeError as e:
            log.debug(f"Could not parse HOPS JSON: '{f['fn']}'")
            log.debug(e)
            return None

        # Convert JSON to dict for easier manipulation
        for s in parsed_json:
            s_name = self.clean_s_name(s, f)
            if s_name in self.hops_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name=s_name)
            self.hops_data[s_name] = {}
            for t in parsed_json[s]:
                self.hops_data[s_name][t] = parsed_json[s][t]

    def hops_heatmap(self):
        """Heatmap showing all statuses for every sample"""
        samples = []
        for s in self.hops_data:
            samples.append(s)

        # As all samples always have same taxa, will take from the first sample
        taxa = []
        for t in self.hops_data[samples[0]]:
            taxa.append(t.replace("_", " "))

        # Get values from named list into a list of lists required for heatmap
        data = []
        for s, d in self.hops_data.items():
            row = []
            for val in d.values():
                # Values can be lists of 1 element, so flattening to single value
                if isinstance(val, list):
                    val = val[0]
                row.append(val)
            data.append(row)

        pconfig = {
            "id": "hops-heatmap",
            "title": "HOPS: Potential Candidates",
            "xTitle": "Node",
            "yTitle": "Sample",
            "square": False,
            "colstops": [
                [1, "#ededed"],
                [2, "#FFFFC5"],
                [3, "#F2B26C"],
                [4, "#AD2A2B"],
            ],
            "decimalPlaces": 0,
            "legend": True,
            "datalabels": False,
            "xcats_samples": False,
        }

        extra_warning = ""
        if len(self.hops_data) > 20:
            extra_warning = """
            <div class="alert alert-warning">
                Large numbers of samples can result in Y-axis labels
                overlapping. Drag the handle at the bottom of the plot down
                to expand and see all samples names.
            </div>
                """

        self.add_section(
            name="Potential Candidates",
            anchor="hops_heatmap",
            description="""
            Heatmap of candidate taxa for downstream aDNA analysis, with
            intensity representing additive categories of possible 'positive'
            hits.
            """
            + extra_warning,
            helptext="""
            HOPS assigns a category based on how many ancient DNA
            characteristics a given node (i.e. taxon) in a sample has.
            The colours indicate the following:

            * <span style="background-color: #ededed; padding:0.2rem 1rem;">**Grey**</span> - No characteristics detected
            * <span style="background-color: #FFFFC5; padding:0.2rem 1rem;">**Yellow**</span> - Small edit distance from reference
            * <span style="background-color: #F2B26C; padding:0.2rem 1rem;">**Orange**</span> - Typical aDNA damage pattern
            * <span style="background-color: #AD2a2B; padding:0.2rem 1rem;">**Red**</span> - Small edit distance _and_ aDNA damage pattern

            A red category typically indicates a good candidate for further investigation
            in downstream analysis.
            """,
            plot=heatmap.plot(data, xcats=taxa, ycats=samples, pconfig=pconfig),
        )



# -*- coding: utf-8 -*-

""" MultiQC module to parse output files from VarScan2 """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="VarScan2",
            anchor="varscan",
            href="http://dkoboldt.github.io/varscan/",
            info="variant detection in massively parallel sequencing data",
            doi=["10.1101/gr.129684.111", "10.1093/bioinformatics/btp373"],
        )

        # Find and load VarScan2 reports - there are 3 different ones, but all with identical content (differentiated by header)
        self.varscan2_data = dict()
        for f in self.find_log_files("varscan2/mpileup2snp", filehandles=True):
            parsed_data = self.parse_varscan(f)
            s_name = self.clean_s_name(parsed_data["sample_name"], f)
            # Drop existing sample_name from dict now
            del parsed_data["sample_name"]
            if parsed_data is not None and len(parsed_data) > 0:
                self.varscan2_data[s_name] = parsed_data
                self.add_data_source(f, s_name, section="mpileup2snp")

        for f in self.find_log_files("varscan2/mpileup2indel", filehandles=True):
            parsed_data = self.parse_varscan(f)
            s_name = self.clean_s_name(parsed_data["sample_name"], f)
            # Drop existing sample_name from dict now
            del parsed_data["sample_name"]
            if parsed_data is not None and len(parsed_data) > 0:
                self.varscan2_data[s_name] = parsed_data
                self.add_data_source(f, s_name, section="mpileup2indel")

        for f in self.find_log_files("varscan2/mpileup2cns", filehandles=True):
            parsed_data = self.parse_varscan(f)
            s_name = self.clean_s_name(parsed_data["sample_name"], f)
            # Drop existing sample_name from dict now
            del parsed_data["sample_name"]
            if parsed_data is not None and len(parsed_data) > 0:
                self.varscan2_data[s_name] = parsed_data
                self.add_data_source(f, s_name, section="mpileup2cns")

        # Filter to strip out ignored sample names
        self.varscan2_data = self.ignore_samples(self.varscan2_data)

        # Warning when no files are found
        if len(self.varscan2_data) == 0:
            raise ModuleNoSamplesFound

        # Write parsed data to a file
        self.write_data_file(self.varscan2_data, "multiqc_varscan2_summary")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Found reports or not?
        log.info(f"Found {len(self.varscan2_data)} reports")

        # Basic Stats Table
        self.varscan2_general_stats_table()

        # Basic barplot section
        self.varscan2_counts_barplot()

    # Varscan2 reports in SNP mode only snps, in indel only indel and in CNS all variants found
    # Total variants = SNPs + Indels
    def parse_varscan(self, f):
        """Parse a VarScan2 report"""
        parsed = dict()
        regexes = {
            "sample_name": r"(?:Reading input from )(\w+.+)",
            "min_coverage": r"(?:Min coverage:)\s(\d+)",
            "min_reads2": r"(?:Min reads2:)\s(\d+)",
            "min_var_freq": r"(?:Min var freq:)\s(\d+\.\d+)",
            "min_avg_qual": r"(?:Min avg qual:)\s(\d+)",
            "p_val_threshold": r"(?:P-value thresh:)\s(\d+\.\d+)",
            "bases_in_pileup": r"(\d+)(?:\sbases in pileup file)",
            "variant_total": r"(\d+)(?:\svariant positions \()",
            "variant_snps": r"(?:\d+)(?:\svariant positions \()(\d+)",
            "variant_indels": r"(?:\d+)(?:\svariant positions \()(?:\d+ SNP, )(\d+)",
            "failed_strand_filter": r"(\d+)(?:\swere failed by the strand-filter)",
            "variant_reported_total": r"(\d+)(?:\svariant positions reported)",
            "variant_reported_snps": r"(?:\d+)(?:\svariant positions reported \()(\d+)",
            "variant_reported_indels": r"(?:\d+)(?:\svariant positions reported \()(?:\d+ SNP, )(\d+)",
        }
        for line in f["f"]:
            # Search regexes for stats
            for k, r in regexes.items():
                match = re.search(r, line)
                if match:
                    if k not in ["sample_name", "p_val_threshold", "min_var_freq"]:
                        parsed[k] = int(match.group(1))
                    if k == "sample_name":
                        parsed[k] = match.group(1)
                    if k in ["p_val_threshold", "min_var_freq"]:
                        parsed[k] = float(match.group(1))

        # The failed_strand_filter combines the different classes, so recompute
        if "variant_total" in parsed and "variant_reported_total" in parsed:
            parsed["variant_total_failed"] = parsed["variant_total"] - parsed["variant_reported_total"]
        if "variant_snps" in parsed and "variant_reported_snps" in parsed:
            parsed["variant_snps_failed"] = parsed["variant_snps"] - parsed["variant_reported_snps"]
        if "variant_indels" in parsed and "variant_reported_indels" in parsed:
            parsed["variant_indels_failed"] = parsed["variant_indels"] - parsed["variant_reported_indels"]

        return parsed

    # Add to general stats table
    def varscan2_general_stats_table(self):
        """Take the parsed stats from the VarScan2 report and add it to the
        General Statistics table
        """

        headers = {
            "variant_reported_snps": {
                "title": "SNPs reported",
                "description": "Total number of SNPs reported.",
                "min": 0,
                "scale": "Spectral",
                "format": "{:,.0f}",
                "shared_key": "snps",
            },
            "variant_reported_indels": {
                "title": "INDELs reported",
                "description": "Total number INDELs reported.",
                "min": 0,
                "scale": "Spectral",
                "format": "{:,.0f}",
                "shared_key": "indels",
            },
            "variant_reported_total": {
                "title": "Variants reported",
                "description": "Total number of variants reported.",
                "min": 0,
                "scale": "Spectral",
                "format": "{:,.0f}",
                "shared_key": "variants",
                "hidden": True,
            },
            "variant_snps": {
                "title": "SNPs",
                "description": "Total number of SNPs detected",
                "min": 0,
                "scale": "BrBG",
                "format": "{:,.0f}",
                "shared_key": "snps",
                "hidden": True,
            },
            "variant_indels": {
                "title": "INDELs",
                "description": "Total number of INDELs detected",
                "min": 0,
                "scale": "BrBG",
                "format": "{:,.0f}",
                "shared_key": "indels",
                "hidden": True,
            },
            "variant_total": {
                "title": "Variants",
                "description": "Total number of variants detected",
                "min": 0,
                "scale": "BrBG",
                "format": "{:,.0f}",
                "shared_key": "variants",
                "hidden": True,
            },
            "failed_strand_filter": {
                "title": "Failed Strand Filter",
                "description": "Total number variants failing the strand-filter.",
                "min": 0,
                "scale": "YlOrBr",
                "format": "{:,.0f}",
                "shared_key": "variants",
                "hidden": True,
            },
            "bases_in_pileup": {
                "title": f"{config.base_count_prefix} Bases in Pileup",
                "description": "Number of bases in pileup input for VarScan2 ()",
                "scale": "Greens",
                "modify": lambda x: x * config.base_count_multiplier,
                "shared_key": "base_count",
                "hidden": True,
            },
        }

        self.general_stats_addcols(self.varscan2_data, headers)

    def varscan2_counts_barplot(self):
        """Make the HighCharts HTML to plot the reported SNPs"""
        # 146 variant positions (106 SNP, 40 indel)
        # 12 were failed by the strand-filter
        # 99 variant positions reported (99 SNP, 0 indel)

        # Specify the order of the different possible categories
        cats = [
            {
                "variant_reported_snps": {"name": "SNPs reported", "color": "#7fc97f"},
                "variant_snps_failed": {"name": "SNPs not reported", "color": "#fb8072"},
            },
            {
                "variant_reported_indels": {"name": "INDELs reported", "color": "#7fc97f"},
                "variant_indels_failed": {"name": "INDELs not reported", "color": "#fb8072"},
            },
        ]
        # Config for the plot
        pconfig = {
            "id": "varscan2_variant_counts_plot",
            "title": "VarScan2: Variants detected",
            "ylab": "Number of SNPs",
            "cpswitch_counts_label": "Number of Variants",
            "hide_zero_cats": False,
            "data_labels": [{"name": "SNPs", "ylab": "Number of SNPs"}, {"name": "INDELs", "ylab": "Number of INDELs"}],
        }

        self.add_section(
            name="Variants detected",
            anchor="varscan2_variant_counts",
            description="""
            This plot shows the total number of variant positions, broken down by those that were reported or not.
            """,
            plot=bargraph.plot([self.varscan2_data, self.varscan2_data], cats, pconfig),
        )



""" MultiQC module to parse output from Adapter Removal """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"AdapterRemoval ver. ([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Adapter Removal",
            anchor="adapterRemoval",
            target="Adapter Removal",
            href="https://github.com/MikkelSchubert/adapterremoval",
            info=" rapid adapter trimming, identification, and read merging ",
            doi=["10.1186/s13104-016-1900-2", "10.1186/1756-0500-5-337"],
        )

        self.__read_type = None
        self.__any_paired = False
        self.__collapsed = None
        self.__any_collapsed = False
        self.s_name = None
        self.adapter_removal_data = {}

        self.len_dist_plot_data = {
            "mate1": dict(),
            "mate2": dict(),
            "singleton": dict(),
            "collapsed": dict(),
            "collapsed_truncated": dict(),
            "discarded": dict(),
            "all": dict(),
        }

        parsed_data = None
        for f in self.find_log_files("adapterRemoval", filehandles=True):
            self.s_name = f["s_name"]
            try:
                parsed_data = self.parse_settings_file(f)
            except ModuleNoSamplesFound:
                continue
            if parsed_data is not None:
                self.adapter_removal_data[self.s_name] = parsed_data
                self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.adapter_removal_data = self.ignore_samples(self.adapter_removal_data)

        if len(self.adapter_removal_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.adapter_removal_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.adapter_removal_data, "multiqc_adapter_removal")

        # add data to Basic Stats Table
        self.adapter_removal_stats_table()

        self.adapter_removal_retained_chart()
        self.adapter_removal_length_dist_plot()

    def parse_settings_file(self, f):
        self.result_data = {
            "total": None,
            "unaligned": None,
            "aligned": None,
            "reads_total": None,
            "retained": None,
            "percent_aligned": None,
        }

        settings_data = {"header": []}

        block_title = None
        for i, line in enumerate(f["f"]):
            line = line.rstrip("\n")
            if line == "":
                continue

            if line.startswith("AdapterRemoval"):
                version_match = re.search(VERSION_REGEX, line)
                if version_match:
                    self.add_software_version(version_match.group(1), self.s_name)

            if not block_title:
                block_title = "header"
                settings_data[block_title].append(str(line))
                continue

            if line.startswith("["):
                block_title = str(line.strip("[]"))
                settings_data[block_title] = []
                continue

            settings_data[block_title].append(str(line))

        # set data for further working
        self.set_result_data(settings_data)

        return self.result_data

    def set_result_data(self, settings_data):
        # set read and collapsed type
        self.set_ar_type(settings_data["Length distribution"])

        # set result_data
        self.set_trim_stat(settings_data["Trimming statistics"])
        self.set_len_dist(settings_data["Length distribution"])

    def set_ar_type(self, len_dist_data):
        head_line = len_dist_data[0].rstrip("\n").split("\t")

        self.__read_type = "paired" if head_line[2] == "Mate2" else "single"
        if not self.__any_paired:
            self.__any_paired = True if head_line[2] == "Mate2" else False

        self.__collapsed = True if head_line[-3] == "CollapsedTruncated" else False
        if not self.__any_collapsed:
            self.__any_collapsed = True if head_line[-3] == "CollapsedTruncated" else False

        # biological/technical relevance is not clear -> skip
        if self.__read_type == "single" and self.__collapsed:
            log.warning(f"Case single-end and collapse is not implemented -> File {self.s_name} skipped")
            raise ModuleNoSamplesFound

    def set_trim_stat(self, trim_data):
        required = [
            "total",
            "unaligned",
            "aligned",
            "discarded_m1",
            "singleton_m1",
            "retained",
            "discarded_m2",
            "singleton_m2",
            "full-length_cp",
            "truncated_cp",
        ]
        data_pattern = {"total": 0, "unaligned": 1, "aligned": 2, "discarded_m1": 3, "singleton_m1": 4, "retained": 6}

        if self.__read_type == "paired":
            data_pattern["discarded_m2"] = 5
            data_pattern["singleton_m2"] = 6
            if not self.__collapsed:
                data_pattern["retained"] = 8
            else:
                data_pattern["full-length_cp"] = 8
                data_pattern["truncated_cp"] = 9
                data_pattern["retained"] = 10

        for field in required:
            if field in data_pattern:
                tmp = trim_data[data_pattern[field]]
                value = tmp.split(": ")[1]
                self.result_data[field] = int(value)
            else:
                self.result_data[field] = 0

        reads_total = self.result_data["total"]
        aligned_total = self.result_data["aligned"]
        unaligned_total = self.result_data["unaligned"]
        if self.__read_type == "paired":
            reads_total = self.result_data["total"] * 2
            aligned_total = self.result_data["aligned"] * 2
            unaligned_total = self.result_data["unaligned"] * 2

        self.result_data["aligned_total"] = aligned_total
        self.result_data["unaligned_total"] = unaligned_total
        self.result_data["reads_total"] = reads_total
        self.result_data["discarded_total"] = reads_total - self.result_data["retained"]
        if self.__read_type == "paired":
            if not self.__collapsed:
                self.result_data["paired_reads"] = (
                    self.result_data["retained"] - self.result_data["singleton_m1"] - self.result_data["singleton_m2"]
                )
            else:
                self.result_data["paired_reads"] = (
                    self.result_data["retained"]
                    - self.result_data["full-length_cp"]
                    - self.result_data["truncated_cp"]
                    - self.result_data["singleton_m1"]
                    - self.result_data["singleton_m2"]
                )
            full_length_cp = self.result_data["full-length_cp"] * 2
            truncated_cp = self.result_data["truncated_cp"] * 2
            self.result_data["full-length_cp"] = full_length_cp
            self.result_data["truncated_cp"] = truncated_cp
        try:
            self.result_data["percent_aligned"] = (
                float(self.result_data["aligned"]) * 100.0 / float(self.result_data["total"])
            )
        except ZeroDivisionError:
            self.result_data["percent_aligned"] = 0
        if self.__collapsed:
            try:
                self.result_data["percent_collapsed"] = (
                    float(self.result_data["full-length_cp"] + self.result_data["truncated_cp"])
                    * 100.0
                    / float(self.result_data["reads_total"])
                )
            except ZeroDivisionError:
                self.result_data["percent_collapsed"] = 0
        try:
            self.result_data["percent_discarded"] = (
                float(self.result_data["discarded_m1"] + self.result_data["discarded_m2"])
                * 100.0
                / float(self.result_data["reads_total"])
            )
        except ZeroDivisionError:
            self.result_data["percent_discarded"] = 0

    def set_len_dist(self, len_dist_data):
        for line in len_dist_data[1:]:
            l_data = line.rstrip("\n").split("\t")
            l_data = list(map(int, l_data))

            # initialize file name
            if self.s_name not in self.len_dist_plot_data["mate1"]:
                self.len_dist_plot_data["mate1"][self.s_name] = dict()
                self.len_dist_plot_data["mate2"][self.s_name] = dict()
                self.len_dist_plot_data["singleton"][self.s_name] = dict()
                self.len_dist_plot_data["collapsed"][self.s_name] = dict()
                self.len_dist_plot_data["collapsed_truncated"][self.s_name] = dict()
                self.len_dist_plot_data["discarded"][self.s_name] = dict()
                self.len_dist_plot_data["all"][self.s_name] = dict()

            if self.__read_type == "single":
                if not self.__collapsed:
                    self.len_dist_plot_data["mate1"][self.s_name][l_data[0]] = l_data[1]
                    self.len_dist_plot_data["mate2"][self.s_name][l_data[0]] = None
                    self.len_dist_plot_data["singleton"][self.s_name][l_data[0]] = None
                    self.len_dist_plot_data["collapsed"][self.s_name][l_data[0]] = None
                    self.len_dist_plot_data["collapsed_truncated"][self.s_name][l_data[0]] = None
                    self.len_dist_plot_data["discarded"][self.s_name][l_data[0]] = l_data[2]
                    self.len_dist_plot_data["all"][self.s_name][l_data[0]] = l_data[3]
                else:
                    # this case should not be reached (see case at method set_ar_type())
                    pass
            else:
                if not self.__collapsed:
                    self.len_dist_plot_data["mate1"][self.s_name][l_data[0]] = l_data[1]
                    self.len_dist_plot_data["mate2"][self.s_name][l_data[0]] = l_data[2]
                    self.len_dist_plot_data["singleton"][self.s_name][l_data[0]] = l_data[3]
                    self.len_dist_plot_data["collapsed"][self.s_name][l_data[0]] = None
                    self.len_dist_plot_data["collapsed_truncated"][self.s_name][l_data[0]] = None
                    self.len_dist_plot_data["discarded"][self.s_name][l_data[0]] = l_data[4]
                    self.len_dist_plot_data["all"][self.s_name][l_data[0]] = l_data[5]
                else:
                    self.len_dist_plot_data["mate1"][self.s_name][l_data[0]] = l_data[1]
                    self.len_dist_plot_data["mate2"][self.s_name][l_data[0]] = l_data[2]
                    self.len_dist_plot_data["singleton"][self.s_name][l_data[0]] = l_data[3]
                    self.len_dist_plot_data["collapsed"][self.s_name][l_data[0]] = l_data[4]
                    self.len_dist_plot_data["collapsed_truncated"][self.s_name][l_data[0]] = l_data[5]
                    self.len_dist_plot_data["discarded"][self.s_name][l_data[0]] = l_data[6]
                    self.len_dist_plot_data["all"][self.s_name][l_data[0]] = l_data[7]

    def adapter_removal_stats_table(self):
        headers = {
            "percent_aligned": {
                "title": "% Trimmed",
                "description": "% trimmed reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn-rev",
                "shared_key": "percent_aligned",
            },
            "aligned_total": {
                "title": f"{config.read_count_prefix} Reads Trimmed",
                "description": f"Total trimmed reads ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "min": 0,
                "scale": "PuBu",
                "shared_key": "read_count",
            },
        }
        if self.__any_collapsed:
            headers["percent_collapsed"] = {
                "title": "% Collapsed",
                "description": "% collapsed reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn-rev",
                "shared_key": "percent_aligned",
            }
        headers["percent_discarded"] = {
            "title": "% Discarded",
            "description": "% discarded reads",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "RdYlGn-rev",
            "shared_key": "percent_discarded",
        }
        self.general_stats_addcols(self.adapter_removal_data, headers)

    def adapter_removal_retained_chart(self):
        pconfig = {
            "title": "Adapter Removal: Discarded Reads",
            "id": "ar_retained_plot",
            "ylab": "# Reads",
            "hide_zero_cats": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        cats_pec = {}
        if self.__any_paired:
            cats_pec["paired_reads"] = {"name": "Uncollapsed Paired Reads"}

        cats_pec["singleton_m1"] = {"name": "Singleton R1"}

        if self.__any_paired:
            cats_pec["singleton_m2"] = {"name": "Singleton R2"}

            if self.__any_collapsed:
                cats_pec["full-length_cp"] = {"name": "Full-length Collapsed Reads"}
                cats_pec["truncated_cp"] = {"name": "Truncated Collapsed Reads"}

        cats_pec["discarded_m1"] = {"name": "Discarded R1"}

        if self.__any_paired:
            cats_pec["discarded_m2"] = {"name": "Discarded R2"}
        if self.__any_collapsed:
            retained_chart_description = "The number of input sequences that were retained, collapsed, and discarded. Be aware that the number of collapsed reads in the output FASTQ will be half of the numbers displayed in this plot, because both R1 and R2 of the collapsed sequences are counted here."
        else:
            retained_chart_description = "The number of input sequences that were retained and discarded."
        self.add_section(
            name="Retained and Discarded Reads",
            anchor="adapter_removal_retained_plot",
            description=retained_chart_description,
            plot=bargraph.plot(self.adapter_removal_data, cats_pec, pconfig),
        )

    def adapter_removal_length_dist_plot(self):
        pconfig = {
            "title": "Adapter Removal: Length Distribution",
            "id": "ar_length_count_plot",
            "ylab": "Counts",
            "xlab": "read length",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>{point.x} bp trimmed</b>: {point.y:.0f}",
            "data_labels": None,
        }

        lineplot_data = [self.len_dist_plot_data["all"], self.len_dist_plot_data["mate1"]]
        data_labels = [
            {"name": "All", "ylab": "Count"},
            {"name": "Mate1", "ylab": "Count"},
        ]
        if self.__any_paired:
            lineplot_data.extend([self.len_dist_plot_data["mate2"], self.len_dist_plot_data["singleton"]])
            data_labels.extend(
                [
                    {"name": "Mate2", "ylab": "Count"},
                    {"name": "Singleton", "ylab": "Count"},
                ]
            )
            if self.__any_collapsed:
                lineplot_data.extend(
                    [self.len_dist_plot_data["collapsed"], self.len_dist_plot_data["collapsed_truncated"]]
                )
                data_labels.extend(
                    [{"name": "Collapsed", "ylab": "Count"}, {"name": "Collapsed Truncated", "ylab": "Count"}]
                )
        lineplot_data.append(self.len_dist_plot_data["discarded"])
        data_labels.append({"name": "Discarded", "ylab": "Count"})

        pconfig["data_labels"] = data_labels

        self.add_section(
            name="Length Distribution",
            anchor="ar_length_count",
            description="The length distribution of reads after processing adapter alignment.",
            plot=linegraph.plot(lineplot_data, pconfig),
        )



""" MultiQC submodule to parse output from Picard IlluminaBasecallingMetrics """

import logging

from multiqc.modules.picard import util
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard IlluminaBasecallingMetrics reports and parse their data"""

    data_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/collectilluminabasecallingmetrics", filehandles=True):
        keys = None

        for line in f["f"]:
            if util.is_line_right_before_table(line, "IlluminaBasecallingMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")

            elif keys:
                vals = line.strip("\n").split("\t")
                if len(vals) != len(keys):
                    keys = None
                    continue

                data = dict(zip(keys, vals))
                # We only care about the last line
                if data["MOLECULAR_BARCODE_SEQUENCE_1"].strip() == "":
                    data.pop("MOLECULAR_BARCODE_SEQUENCE_1")
                    data.pop("MOLECULAR_BARCODE_NAME")
                    s_name = data["LANE"]
                    if s_name in data_by_sample:
                        log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                    module.add_data_source(f, s_name=s_name, section="IlluminaBasecallingMetrics")
                    data_by_sample[s_name] = data

    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, f"multiqc_{module.anchor}_IlluminaBasecallingMetrics")

    module.add_section(
        name="Basecalling Metrics",
        anchor=f"{module.anchor}-illuminabasecallingmetrics",
        description="Quality control metrics for each lane of an Illumina flowcell.",
        helptext="""
        For full details, please see the [Picard Documentation](http://broadinstitute.github.io/picard/picard-metric-definitions.html#IlluminaBasecallingMetrics).

        * `PF_BASES` / `NPF_BASES` :  The total number of passing-filter / not-passing-filter bases assigned to the index.
        * `PF_READS` / `NPF_READS` :  The total number of passing-filter / not-passing-filter reads assigned to the index.
        * `PF_CLUSTERS` / `NPF_CLUSTERS` :  The total number of passing-filter / not-passing-filter clusters assigned to the index.

        `NPF` stands for _"not passing filter"_ and is calculated by subtracting the `PF_` metric from the `TOTAL_` Picard metrics.
        """,
        plot=lane_metrics_plot(module, data_by_sample),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)


def lane_metrics_table(self, data):
    headers = {
        "TOTAL_BASES": {"title": "Total Bases"},
        "PF_BASES": {"title": "Passing Filter Bases"},
        "TOTAL_READS": {"title": "Total Reads"},
        "PF_READS": {"title": "Passing Filter Reads"},
        "TOTAL_CLUSTERS": {"title": "Total Cluster"},
        "PF_CLUSTERS": {"title": "Passing Filter Clusters"},
    }

    table_config = {
        "id": f"{self.anchor}-illumina-basecalling-metrics-table",
        "namespace": self.name,
        "table_title": f"{self.name} Illumina Base Calling Metrics",
    }
    tdata = {}
    for lane_number, lane in data.items():
        tdata[f"L{lane_number}"] = lane
    return table.plot(tdata, headers, table_config)


def lane_metrics_plot(self, data):
    plot_config = {
        "id": f"plot-{self.anchor}-illuminabasecallingmetrics",
        "title": f"{self.name}: Illumina Basecalling Metrics",
        "ylab": "Lane",
        "data_labels": [
            {"name": "Bases", "ylab": "Number of Bases"},
            {"name": "Reads", "ylab": "Number of Reads"},
            {"name": "Clusters", "ylab": "Number of Clusters"},
        ],
    }

    plot_cats = [
        {
            "PF_BASES": {"title": "Passing Filter Bases"},
            "NPF_BASES": {"title": "Non Passing Filter Bases"},
        },
        {
            "PF_READS": {"title": "Passing Filter Reads"},
            "NPF_READS": {"title": "Non Passing Filter Reads"},
        },
        {
            "PF_CLUSTERS": {"title": "Passing Filter Clusters"},
            "NPF_CLUSTERS": {"title": "Non Passing Filter Clusters"},
        },
    ]
    tdata = {}
    for lane_number, lane in data.items():
        lane["NPF_BASES"] = int(lane["TOTAL_BASES"]) - int(lane["PF_BASES"])
        lane["NPF_READS"] = int(lane["TOTAL_READS"]) - int(lane["PF_READS"])
        lane["NPF_CLUSTERS"] = int(lane["TOTAL_CLUSTERS"]) - int(lane["PF_CLUSTERS"])
        tdata[f"L{lane_number}"] = lane
    return bargraph.plot([tdata, tdata, tdata], plot_cats, plot_config)



""" MultiQC module to parse output from SexdetErrmine """


import json
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, scatter

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """SexDeterrmine module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="SexDetErrmine",
            anchor="sexdeterrmine",
            href="https://github.com/TCLamnidis/Sex.DetERRmine",
            info="""A python script to calculate the relative coverage of X and Y chromosomes,
            and their associated error bars, from the depth of coverage at specified SNPs.""",
            doi="10.1038/s41467-018-07483-5",
        )

        # Find and load any DeDup reports
        self.sexdet_data = dict()

        # Find and load JSON file
        for f in self.find_log_files("sexdeterrmine", filehandles=True):
            self.parse_data(f)

        # Filter samples
        self.sexdet_data = self.ignore_samples(self.sexdet_data)

        # Return if no samples found
        if len(self.sexdet_data) == 0:
            raise ModuleNoSamplesFound

        # Save data output file
        self.write_data_file(self.sexdet_data, "multiqc_sexdeter_metrics")

        # Add to General Statistics
        self.addSummaryMetrics()

        # Plots
        self.snp_rate_scatterplot()
        self.read_count_barplot()
        self.snp_count_barplot()

    def parse_data(self, f):
        try:
            data = json.load(f["f"])
        except Exception as e:
            log.debug(e)
            log.warning(f"Could not parse SexDeterrmine JSON: '{f['fn']}'")
            return

        # Get the version
        version = str(data["Metadata"]["version"])

        # Parse JSON data to a dict
        for s_name in data:
            if s_name == "Metadata":
                continue

            s_clean = self.clean_s_name(s_name, f)
            if s_clean in self.sexdet_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_clean}")

            self.add_data_source(f, s_clean)
            self.add_software_version(version, s_clean)
            self.sexdet_data[s_clean] = dict()

            for k, v in data[s_name].items():
                try:
                    self.sexdet_data[s_clean][k] = float(v)
                except ValueError:
                    self.sexdet_data[s_clean][k] = v

    def addSummaryMetrics(self):
        """Take the parsed stats from SexDetErrmine and add it to the main plot"""

        headers = {
            "RateErrX": {
                "title": "Err Rate X",
                "description": "Rate of Error for Chr X",
                "scale": "OrRd",
                "hidden": True,
                "shared_key": "snp_err_rate",
            },
            "RateErrY": {
                "title": "Err Rate Y",
                "description": "Rate of Error for Chr Y",
                "scale": "OrRd",
                "hidden": True,
                "shared_key": "snp_err_rate",
            },
            "RateX": {
                "title": "Rate X",
                "description": "Number of positions on Chromosome X vs Autosomal positions.",
                "scale": "PuBuGn",
                "shared_key": "snp_count",
            },
            "RateY": {
                "title": "Rate Y",
                "description": "Number of positions on Chromosome Y vs Autosomal positions.",
                "scale": "BuPu",
                "shared_key": "snp_count",
            },
        }

        self.general_stats_addcols(self.sexdet_data, headers)

    def read_count_barplot(self):
        """Make a bar plot showing read counts on Autosomal, X and Y chr"""
        cats = {
            "NR Aut": {"name": "Autosomal Reads"},
            "NrX": {"name": "Reads on X"},
            "NrY": {"name": "Reads on Y"},
        }
        config = {
            "id": "sexdeterrmine-readcounts-plot",
            "title": "SexDetErrmine: Read Counts",
            "ylab": "# Reads",
        }

        self.add_section(
            name="Read Counts",
            anchor="sexdeterrmine-readcounts",
            description="The number of reads covering positions on the autosomes, X and Y chromosomes.",
            plot=bargraph.plot(self.sexdet_data, cats, config),
        )

    def snp_rate_scatterplot(self):
        """Make a scatter plot showing relative coverage on X and Y chr"""
        data = {}
        for sample in self.sexdet_data:
            try:
                data[sample] = {"x": self.sexdet_data[sample]["RateX"], "y": self.sexdet_data[sample]["RateY"]}
            except KeyError:
                pass

        config = {
            "id": "sexdeterrmine-rate-plot",
            "title": "SexDetErrmine: Relative coverage",
            "ylab": "Relative Cov. on Y",
            "xlab": "Relative Cov. on X",
        }

        if len(data) > 0:
            self.add_section(
                name="Relative Coverage",
                anchor="sexdeterrmine-rates",
                description="The coverage on the X vs Y chromosome, relative to coverage on the Autosomes.",
                helptext="""
                Males are expected to have a roughly equal X- and Y-rates, while females are expected to have a Y-rate of 0 and an X-rate of 1.
                Placement between the two clusters can be indicative of contamination, while placement with higher than expected X- and/or Y-rates can be indicative of sex chromosome aneuploidy.
                """,
                plot=scatter.plot(data, config),
            )

    def snp_count_barplot(self):
        """Make a bar plot showing read counts on Autosomal, X and Y chr"""
        cats = {
            "Snps Autosomal": {"name": "Autosomal SNPs"},
            "XSnps": {"name": "SNPs on X"},
            "YSnps": {"name": "SNPs on Y"},
        }

        config = {
            "id": "sexdeterrmine-snps-plot",
            "title": "SexDetErrmine: SNP Counts",
            "ylab": "# Reads",
        }

        self.add_section(
            name="SNP Counts",
            anchor="sexdeterrmine-snps",
            description="Total number of SNP positions. When supplied with a BED file, this includes only positions specified there.",
            plot=bargraph.plot(self.sexdet_data, cats, config),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
multiqc.multiqc
~~~~~~~~~~~~~~~~~~~~~
The main function to run MultiQC. Sorry about the messy namespace.
Primarily called by multiqc.__main__.py
Imported by __init__.py so available as multiqc.run()
"""
import base64
import errno
import io
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import json
import traceback

import jinja2
import requests
import rich
import rich_click as click
from packaging import version
from rich.syntax import Syntax

from .modules.base_module import ModuleNoSamplesFound
from .plots import table
from .utils import config, log, megaqc, plugin_hooks, report, software_versions, strict_helpers, util_functions
from .utils.util_functions import strtobool

# Set up logging
start_execution_time = time.time()
logger = config.logger

OLDEST_SUPPORTED_PYTHON_VERSION = "3.8"


# Configuration for rich-click CLI help
click.rich_click.USE_RICH_MARKUP = True
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.HEADER_TEXT = f"[dark_orange]///[/] [bold][link=https://multiqc.info]MultiQC[/link][/] :{util_functions.choose_emoji()}: [dim]| v{config.version}"
click.rich_click.FOOTER_TEXT = "See [link=http://multiqc.info]http://multiqc.info[/] for more details."
click.rich_click.ERRORS_SUGGESTION = f"This is MultiQC [cyan]v{config.version}[/]\nFor more help, run '[yellow]multiqc --help[/]' or visit [link=http://multiqc.info]http://multiqc.info[/]"
click.rich_click.STYLE_ERRORS_SUGGESTION = ""
click.rich_click.OPTION_GROUPS = {
    "multiqc": [
        {
            "name": "Main options",
            "options": [
                "--force",
                "--config",
                "--cl-config",
                "--filename",
                "--outdir",
                "--ignore",
                "--ignore-samples",
                "--ignore-symlinks",
                "--file-list",
            ],
        },
        {
            "name": "Choosing modules to run",
            "options": [
                "--module",
                "--exclude",
            ],
        },
        {
            "name": "Sample handling",
            "options": [
                "--dirs",
                "--dirs-depth",
                "--fullnames",
                "--fn_as_s_name",
                "--rename-sample-names",
                "--replace-names",
            ],
        },
        {
            "name": "Report customisation",
            "options": [
                "--title",
                "--comment",
                "--template",
                "--sample-names",
                "--sample-filters",
                "--custom-css-file",
            ],
        },
        {
            "name": "Output files",
            "options": [
                "--flat",
                "--interactive",
                "--export",
                "--data-dir",
                "--no-data-dir",
                "--data-format",
                "--zip-data-dir",
                "--no-report",
                "--pdf",
            ],
        },
        {
            "name": "MultiQC behaviour",
            "options": [
                "--verbose",
                "--quiet",
                "--strict",
                "--development",
                "--require-logs",
                "--profile-runtime",
                "--no-megaqc-upload",
                "--no-ansi",
                "--version",
                "--help",
            ],
        },
    ],
}


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument("analysis_dir", type=click.Path(exists=True), nargs=-1, required=True, metavar="[ANALYSIS DIRECTORY]")
@click.option("-f", "--force", is_flag=True, help="Overwrite any existing reports")
@click.option("-d", "--dirs", is_flag=True, help="Prepend directory to sample names")
@click.option(
    "-dd",
    "--dirs-depth",
    "dirs_depth",
    type=int,
    help="Prepend [yellow i]n[/] directories to sample names. Negative number to take from start of path.",
)
@click.option(
    "-s",
    "--fullnames",
    "no_clean_sname",
    is_flag=True,
    help="Do not clean the sample names [i](leave as full file name)[/]",
)
@click.option(
    "-i",
    "--title",
    type=str,
    help="Report title. Printed as page header, used for filename if not otherwise specified.",
)
@click.option(
    "-b", "--comment", "report_comment", type=str, help="Custom comment, will be printed at the top of the report."
)
@click.option("-n", "--filename", type=str, help="Report filename. Use '[yellow i]stdout[/]' to print to standard out.")
@click.option("-o", "--outdir", type=str, help="Create report in the specified output directory.")
@click.option(
    "-t", "--template", type=click.Choice(config.avail_templates), metavar=None, help="Report template to use."
)
@click.option("-x", "--ignore", type=str, multiple=True, metavar="GLOB EXPRESSION", help="Ignore analysis files")
@click.option(
    "--ignore-samples", "ignore_samples", type=str, multiple=True, metavar="GLOB EXPRESSION", help="Ignore sample names"
)
@click.option("--ignore-symlinks", "ignore_symlinks", is_flag=True, help="Ignore symlinked directories and files")
@click.option(
    "--fn_as_s_name", "use_filename_as_sample_name", is_flag=True, help="Use the log filename as the sample name"
)
@click.option(
    "--replace-names",
    "replace_names",
    type=click.Path(exists=True, readable=True),
    help="TSV file to rename sample names during report generation",
)
@click.option(
    "--sample-names",
    "sample_names",
    type=click.Path(exists=True, readable=True),
    help="TSV file containing alternative sample names for renaming buttons in the report",
)
@click.option(
    "--sample-filters",
    "sample_filters",
    type=click.Path(exists=True, readable=True),
    help="TSV file containing show/hide patterns for the report",
)
@click.option(
    "-l", "--file-list", is_flag=True, help="Supply a file containing a list of file paths to be searched, one per row"
)
@click.option(
    "-e",
    "--exclude",
    metavar="[MODULE NAME]",
    type=click.Choice(sorted(["general_stats"] + list(config.avail_modules.keys()))),
    multiple=True,
    help="Do not use this module. Can specify multiple times.",
)
@click.option(
    "-m",
    "--module",
    metavar="[MODULE NAME]",
    type=click.Choice(sorted(config.avail_modules.keys())),
    multiple=True,
    help="Use only this module. Can specify multiple times.",
)
@click.option(
    "--require-logs",
    "require_logs",
    is_flag=True,
    help="Require all explicitly requested modules to have log files. If not, MultiQC will exit with an error.",
)
@click.option("--data-dir", "make_data_dir", is_flag=True, help="Force the parsed data directory to be created.")
@click.option(
    "--no-data-dir", "no_data_dir", is_flag=True, help="Prevent the parsed data directory from being created."
)
@click.option(
    "-k",
    "--data-format",
    "data_format",
    type=click.Choice(list(config.data_format_extensions.keys())),
    help="Output parsed data in a different format.",
)
@click.option("-z", "--zip-data-dir", "zip_data_dir", is_flag=True, help="Compress the data directory.")
@click.option("--no-report", "no_report", is_flag=True, help="Do not generate a report, only export data and plots")
@click.option(
    "-p", "--export", "export_plots", is_flag=True, help="Export plots as static images in addition to the report"
)
@click.option("-fp", "--flat", "plots_flat", is_flag=True, help="Use only flat plots [i](static images)[/]")
@click.option(
    "-ip",
    "--interactive",
    "plots_interactive",
    is_flag=True,
    help="Use only interactive plots [i](in-browser Javascript)[/]",
)
@click.option(
    "--strict",
    "strict",
    is_flag=True,
    help="Don't catch exceptions, run additional code checks to help development.",
)
@click.option("--lint", "lint", is_flag=True, hidden=True, help="DEPRECATED: use --strict instead")
@click.option(
    "--development",
    "--dev",
    "development",
    is_flag=True,
    help="Development mode. Do not compress and minimise JS, export uncompressed plot data",
)
@click.option(
    "--pdf",
    "make_pdf",
    is_flag=True,
    help="Creates PDF report with the [i]'simple'[/] template. Requires [link=https://pandoc.org/]Pandoc[/] to be installed.",
)
@click.option(
    "--no-megaqc-upload",
    "no_megaqc_upload",
    is_flag=True,
    help="Don't upload generated report to MegaQC, even if MegaQC options are found",
)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, readable=True),
    multiple=True,
    help="Specific config file to load, after those in MultiQC dir / home dir / working dir.",
)
@click.option("--cl-config", type=str, multiple=True, help="Specify MultiQC config YAML on the command line")
@click.option("-v", "--verbose", count=True, default=0, help="Increase output verbosity.")
@click.option("-q", "--quiet", is_flag=True, help="Only show log warnings")
@click.option("--profile-runtime", is_flag=True, help="Add analysis of how long MultiQC takes to run to the report")
@click.option("--no-ansi", is_flag=True, help="Disable coloured log output")
@click.option(
    "--custom-css-file",
    "custom_css_files",
    type=click.Path(exists=True, readable=True),
    multiple=True,
    help="Custom CSS file to add to the final report",
)
@click.version_option(config.version, prog_name="multiqc")
def run_cli(**kwargs):
    # Main MultiQC run command for use with the click command line, complete with all click function decorators.
    # To make it easy to use MultiQC within notebooks and other locations that don't need click, we simply pass the
    # parsed variables on to a vanilla python function.

    """MultiQC aggregates results from bioinformatics analyses across many samples into a single report.

    It searches a given directory for analysis logs and compiles a HTML report.
    It's a general use tool, perfect for summarising the output from numerous
    bioinformatics tools.

    To run, supply with one or more directory to scan for analysis results.
    For example, to run in the current working directory, use '[blue bold]multiqc .[/]'
    """

    # Pass on to a regular function that can be used easily without click
    multiqc_run = run(**kwargs)

    # End execution using the exit code returned from MultiQC
    sys.exit(multiqc_run["sys_exit_code"])


# Main function that runs MultiQC. Available to use within an interactive Python environment
def run(
    analysis_dir,
    dirs=False,
    dirs_depth=None,
    no_clean_sname=False,
    title=None,
    report_comment=None,
    template=None,
    module=(),
    require_logs=False,
    exclude=(),
    outdir=None,
    ignore=(),
    ignore_samples=(),
    use_filename_as_sample_name=False,
    replace_names=None,
    sample_names=None,
    sample_filters=None,
    file_list=False,
    filename=None,
    make_data_dir=False,
    no_data_dir=False,
    data_format=None,
    zip_data_dir=False,
    force=True,
    ignore_symlinks=False,
    no_report=False,
    export_plots=False,
    plots_flat=False,
    plots_interactive=False,
    strict=False,
    lint=False,  # Deprecated since v1.17
    development=False,
    make_pdf=False,
    no_megaqc_upload=False,
    config_file=(),
    cl_config=(),
    verbose=0,
    quiet=False,
    profile_runtime=False,
    no_ansi=False,
    custom_css_files=(),
    **kwargs,
):
    """MultiQC aggregates results from bioinformatics analyses across many samples into a single report.

    It searches a given directory for analysis logs and compiles a HTML report.
    It's a general use tool, perfect for summarising the output from numerous
    bioinformatics tools.

    To run, supply with one or more directory to scan for analysis results.
    To run here, use 'multiqc .'

    See http://multiqc.info for more details.

    Author: Phil Ewels (http://phil.ewels.co.uk)
    """

    # Set up logging level
    loglevel = log.LEVELS.get(min(verbose, 1), "INFO")
    if quiet:
        loglevel = "WARNING"
        config.quiet = True
    log.init_log(logger, loglevel=loglevel, no_ansi=no_ansi)

    # Throw an error if we are using an unsupported version of Python
    if sys.version_info < tuple(map(int, OLDEST_SUPPORTED_PYTHON_VERSION.split("."))):
        logger.critical(
            "You are running MultiQC with Python {}. "
            "Please upgrade Python! MultiQC does not support Python < {}, "
            "things will break.".format(sys.version_info, OLDEST_SUPPORTED_PYTHON_VERSION)
        )
        return {"report": None, "config": None, "sys_exit_code": 1}

    console = rich.console.Console(
        stderr=True,
        highlight=False,
        force_terminal=util_functions.force_term_colors(),
        color_system=None if no_ansi else "auto",
    )
    console.print(
        f"\n  [dark_orange]///[/] [bold][link=https://multiqc.info]MultiQC[/link][/] :mag: [dim]| v{config.version}\n"
    )
    logger.debug(f"This is MultiQC v{config.version}")

    # Load config files
    plugin_hooks.mqc_trigger("before_config")
    config.mqc_load_userconfig(config_file)
    plugin_hooks.mqc_trigger("config_loaded")

    # Command-line config YAML
    if len(cl_config) > 0:
        config.mqc_cl_config(cl_config)

    report.init()

    # Log the command used to launch MultiQC
    report.multiqc_command = " ".join(sys.argv)
    logger.debug(f"Command used: {report.multiqc_command}")

    # Check that we're running the latest version of MultiQC
    if config.no_version_check is not True:
        try:
            # Fetch the version info from the API
            meta = {
                "version_multiqc": config.short_version,
                "version_python": platform.python_version(),
                "operating_system": platform.system(),
                "is_docker": os.path.exists("/.dockerenv"),
                "is_singularity": os.path.exists("/.singularity.d"),
                "is_conda": os.path.exists(os.path.join(sys.prefix, "conda-meta")),
                "is_ci": strtobool(os.getenv("CI", False)),
            }
            wait_seconds = 2
            try:
                r = requests.get(config.version_check_url, params=meta, timeout=wait_seconds)
            except requests.exceptions.Timeout as e:
                logger.debug(
                    f"Timed out after waiting for {wait_seconds}s for multiqc.info to check latest version: {e}"
                )
            except requests.exceptions.RequestException as e:
                logger.debug(f"Could not connect to multiqc.info for version check: {e}")
            else:
                release_info = r.json()
                # Broadcast log messages if found
                for msg in release_info.get("broadcast_messages", []):
                    if msg.get("message"):
                        level = msg.get("level")
                        if level not in ["debug", "info", "warning", "error", "critical"]:
                            level = "info"
                        getattr(logger, level)(msg["message"])
                # Available update log if newer
                remove_version = version.parse(re.sub(r"[^0-9.]", "", release_info["latest_release"]["version"]))
                this_version = version.parse(re.sub(r"[^0-9.]", "", config.short_version))
                if remove_version > this_version:
                    logger.warning(f"MultiQC Version {release_info['latest_release']['version']} now available!")
                logger.debug(
                    f"Latest MultiQC version is {release_info['latest_release']['version']}, "
                    f"released {release_info['latest_release']['release_date']}"
                )
        except Exception as e:
            logger.debug(f"Could not connect to multiqc.info for version check: {e}")

    # Set up key variables (overwrite config vars from command line)
    if template is not None:
        config.template = template
    if title is not None:
        config.title = title
    if report_comment is not None:
        config.report_comment = report_comment
    if dirs is True:
        config.prepend_dirs = dirs
    if dirs_depth is not None:
        config.prepend_dirs = True
        config.prepend_dirs_depth = dirs_depth
    # Clean up analysis_dir if a string (interactive environment only)
    if isinstance(analysis_dir, str):
        analysis_dir = [analysis_dir]
    config.analysis_dir = analysis_dir
    if outdir is not None:
        config.output_dir = os.path.realpath(outdir)
    if use_filename_as_sample_name:
        config.use_filename_as_sample_name = True
        logger.info("Using log filenames for sample names")
    if make_data_dir:
        config.make_data_dir = True
    if no_data_dir:
        config.make_data_dir = False
    if force:
        config.force = True
    if ignore_symlinks:
        config.ignore_symlinks = True
    if zip_data_dir:
        config.zip_data_dir = True
    if data_format is not None:
        config.data_format = data_format
    if export_plots:
        config.export_plots = True
    if no_report:
        config.make_report = False
    if plots_flat:
        config.plots_force_flat = True
    if plots_interactive:
        config.plots_force_interactive = True
    if lint or config.lint:  # Deprecated since v1.17
        logger.warning(
            "DEPRECIATED: The --lint option is renamed to --strict since MultiQC 1.17. "
            "The old option will be removed in future MultiQC versions, please "
            "update your command line and/or configs."
        )
        strict = True
    if strict:
        config.strict = True
        config.lint = True  # Deprecated since v1.17
        strict_helpers.run_tests()
    if development:
        config.development = True
        config.export_plots = True
        if "png" not in config.export_plot_formats:
            config.export_plot_formats.append("png")
    if make_pdf:
        config.template = "simple"
    if no_megaqc_upload:
        config.megaqc_upload = False
    else:
        config.megaqc_upload = True
    if no_clean_sname:
        config.fn_clean_sample_names = False
        logger.info("Not cleaning sample names")
    if replace_names:
        config.load_replace_names(replace_names)
    if sample_names:
        config.load_sample_names(sample_names)
    config.load_show_hide(sample_filters)
    if len(module) > 0:
        config.run_modules = module
    if len(exclude) > 0:
        config.exclude_modules = exclude
    if require_logs:
        config.require_logs = True
    if profile_runtime:
        config.profile_runtime = True
    if no_ansi:
        config.no_ansi = True
    if custom_css_files:
        config.custom_css_files.extend(custom_css_files)
    config.kwargs = kwargs  # Plugin command line options

    # Discarding the CLI variables to make sure we use only config down below.
    del analysis_dir
    del dirs
    del dirs_depth
    del no_clean_sname
    del title
    del report_comment
    del template
    del module
    del require_logs
    del exclude
    del outdir
    del use_filename_as_sample_name
    del replace_names
    del sample_names
    del sample_filters
    del make_data_dir
    del no_data_dir
    del data_format
    del zip_data_dir
    del force
    del ignore_symlinks
    del no_report
    del export_plots
    del plots_flat
    del plots_interactive
    del strict
    del lint
    del no_megaqc_upload
    del config_file
    del cl_config
    del verbose
    del quiet
    del profile_runtime
    del no_ansi
    del custom_css_files

    plugin_hooks.mqc_trigger("execution_start")

    logger.debug(f"Working dir : {os.getcwd()}")
    if make_pdf:
        logger.info("--pdf specified. Using non-interactive HTML template.")
    logger.debug(f"Template    : {config.template}")
    if config.strict:
        logger.info(
            "Strict mode specified. Will exit early if a module or a template crashed, and will "
            "give warnings if anything is not optimally configured in a module or a template."
        )

    logger.debug("Running Python " + sys.version.replace("\n", " "))

    # Add files if --file-list option is given
    if file_list:
        if len(config.analysis_dir) > 1:
            raise ValueError("If --file-list is given, analysis_dir should have only one plain text file.")
        file_list_path = config.analysis_dir[0]
        config.analysis_dir = []
        with open(file_list_path) as in_handle:
            for line in in_handle:
                if os.path.exists(line.strip()):
                    path = os.path.abspath(line.strip())
                    config.analysis_dir.append(path)
        if len(config.analysis_dir) == 0:
            logger.error(f"No files or directories were added from {file_list_path} using --file-list option.")
            logger.error(f"Please, check that {file_list_path} contains correct paths.")
            raise ValueError("Any files or directories to be searched.")

    if len(ignore) > 0:
        logger.debug(f"Ignoring files, directories and paths that match: {', '.join(ignore)}")
        config.fn_ignore_files.extend(ignore)
        config.fn_ignore_dirs.extend(ignore)
        config.fn_ignore_paths.extend(ignore)
    if len(ignore_samples) > 0:
        logger.debug(f"Ignoring sample names that match: {', '.join(ignore_samples)}")
        config.sample_names_ignore.extend(ignore_samples)
    if filename == "stdout":
        config.output_fn = sys.stdout
        logger.info("Printing report to stdout")
    else:
        if filename is not None and filename.endswith(".html"):
            filename = filename[:-5]
        if filename is None and config.title is not None:
            filename = re.sub(r"[^\w.-]", "", re.sub(r"[-\s]+", "-", config.title)).strip()
            filename += "_multiqc_report"
        if filename is not None:
            if "output_fn_name" not in config.nondefault_config:
                config.output_fn_name = f"{filename}.html"
            if "data_dir_name" not in config.nondefault_config:
                config.data_dir_name = f"{filename}_data"
            if "plots_dir_name" not in config.nondefault_config:
                config.plots_dir_name = f"{filename}_plots"
        if not config.output_fn_name.endswith(".html"):
            config.output_fn_name = f"{config.output_fn_name}.html"

    # Print some status updates
    if config.title is not None:
        logger.info(f"Report title: {config.title}")
    if config.prepend_dirs:
        logger.info("Prepending directory to sample names")

    # Prep module configs
    config.top_modules = [m if isinstance(m, dict) else {m: {}} for m in config.top_modules]
    config.module_order = [m if isinstance(m, dict) else {m: {}} for m in config.module_order]
    mod_keys = [list(m.keys())[0] for m in config.module_order]

    # Lint the module configs
    if config.strict:
        for m in config.avail_modules.keys():
            if m not in mod_keys:
                errmsg = f"LINT: Module '{m}' not found in config.module_order"
                logger.error(errmsg)
                report.lint_errors.append(errmsg)

    # Get the list of modules we want to run, in the order that we want them
    run_modules = [m for m in config.top_modules if list(m.keys())[0] in config.avail_modules.keys()]
    run_modules.extend([{m: {}} for m in config.avail_modules.keys() if m not in mod_keys and m not in run_modules])
    run_modules.extend(
        [
            m
            for m in config.module_order
            if list(m.keys())[0] in config.avail_modules.keys()
            and list(m.keys())[0] not in [list(rm.keys())[0] for rm in run_modules]
        ]
    )

    # Check the module names that were requested to run explicitly
    if len(getattr(config, "run_modules", {})) > 0:
        unknown_modules = [m for m in config.run_modules if m not in config.avail_modules.keys()]
        if unknown_modules:
            logger.error(f"Module(s) in config.run_modules are unknown: {', '.join(unknown_modules)}")
        if len(unknown_modules) == len(config.run_modules):
            logger.critical("No available modules to run!")
            return {"report": report, "config": config, "sys_exit_code": 1}
        config.run_modules = [m for m in config.run_modules if m in config.avail_modules.keys()]
        run_modules = [m for m in run_modules if list(m.keys())[0] in config.run_modules]
        logger.info(f"Only using modules: {', '.join(config.run_modules)}")
    if len(getattr(config, "exclude_modules", {})) > 0:
        logger.info("Excluding modules '{}'".format("', '".join(config.exclude_modules)))
        if "general_stats" in config.exclude_modules:
            config.skip_generalstats = True
            config.exclude_modules = tuple(x for x in config.exclude_modules if x != "general_stats")
        run_modules = [m for m in run_modules if list(m.keys())[0] not in config.exclude_modules]
    if len(run_modules) == 0:
        logger.critical("No analysis modules specified!")
        return {"report": report, "config": config, "sys_exit_code": 1}
    run_module_names = [list(m.keys())[0] for m in run_modules]
    logger.debug(f"Analysing modules: {', '.join(run_module_names)}")

    # Create the temporary working directories
    tmp_dir = tempfile.mkdtemp()
    logger.debug(f"Using temporary directory for creating report: {tmp_dir}")
    config.data_tmp_dir = os.path.join(tmp_dir, "multiqc_data")
    if filename != "stdout" and config.make_data_dir is True:
        config.data_dir = config.data_tmp_dir
        os.makedirs(config.data_dir)
    else:
        config.data_dir = None
    config.plots_tmp_dir = os.path.join(tmp_dir, "multiqc_plots")
    if filename != "stdout" and config.export_plots is True:
        config.plots_dir = config.plots_tmp_dir
        os.makedirs(config.plots_dir)
    else:
        config.plots_dir = None
    if not config.make_report:
        config.output_fn = None

    # Load the template
    template_mod = config.avail_templates[config.template].load()

    # Add an output subdirectory if specified by template
    try:
        config.output_dir = os.path.join(config.output_dir, template_mod.output_subdir)
    except AttributeError:
        pass  # No subdirectory variable given

    # Add custom content section names
    try:
        if "custom_content" in run_module_names:
            run_module_names.extend(config.custom_data.keys())
    except AttributeError:
        pass  # custom_data not in config

    # Always run software_versions module to collect version YAML files
    # Use config.skip_versions_section to exclude from report
    if "software_versions" not in run_module_names:
        run_module_names.append("software_versions")

    # Get the list of files to search
    for d in config.analysis_dir:
        logger.info(f"Search path : {os.path.abspath(d)}")
    report.get_filelist(run_module_names)

    # Only run the modules for which any files were found
    non_empty_modules = {key.split("/")[0].lower() for key, files in report.files.items() if len(files) > 0}
    # Always run custom content, as it can have data purely from a MultiQC config file (no search files)
    if "custom_content" not in non_empty_modules:
        non_empty_modules.add("custom_content")
    run_modules = [m for m in run_modules if list(m.keys())[0].lower() in non_empty_modules]
    run_module_names = [list(m.keys())[0] for m in run_modules]

    if not _required_logs_found(run_module_names):
        return {"report": report, "config": config, "sys_exit_code": 1}

    # Run the modules!
    plugin_hooks.mqc_trigger("before_modules")
    report.modules_output = list()
    sys_exit_code = 0
    total_mods_starttime = time.time()
    for mod_idx, mod_dict in enumerate(run_modules):
        mod_starttime = time.time()
        this_module = list(mod_dict.keys())[0]
        mod_cust_config = list(mod_dict.values())[0]
        if mod_cust_config is None:
            mod_cust_config = {}
        try:
            mod = config.avail_modules[this_module].load()
            mod.mod_cust_config = mod_cust_config  # feels bad doing this, but seems to work
            output = mod()
            if not isinstance(output, list):
                output = [output]
            for m in output:
                report.modules_output.append(m)

            if config.make_report:
                # Copy over css & js files if requested by the theme
                try:
                    for to, path in report.modules_output[-1].css.items():
                        copy_to = os.path.join(tmp_dir, to)
                        os.makedirs(os.path.dirname(copy_to))
                        shutil.copyfile(path, copy_to)
                except OSError as e:
                    if e.errno == errno.EEXIST:
                        pass
                    else:
                        raise
                except AttributeError:
                    pass
                try:
                    for to, path in report.modules_output[-1].js.items():
                        copy_to = os.path.join(tmp_dir, to)
                        os.makedirs(os.path.dirname(copy_to))
                        shutil.copyfile(path, copy_to)
                except OSError as e:
                    if e.errno == errno.EEXIST:
                        pass
                    else:
                        raise
                except AttributeError:
                    pass

        except ModuleNoSamplesFound:
            logger.debug(f"No samples found: {this_module}")
        except UserWarning:  # UserWarning deprecated from 1.16
            msg = f"DEPRECIATED: Please raise 'ModuleNoSamplesFound' instead of 'UserWarning' in module: {this_module}"
            if config.strict:
                logger.error(msg)
                report.lint_errors.append(msg)
            else:
                logger.debug(msg)
            logger.debug(f"No samples found: {this_module}")
        except KeyboardInterrupt:
            shutil.rmtree(tmp_dir)
            logger.critical(
                "User Cancelled Execution!\n{eq}\n{tb}{eq}\n".format(eq=("=" * 60), tb=traceback.format_exc())
                + "User Cancelled Execution!\nExiting MultiQC..."
            )
            sys.exit(1)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive debugging of modules.
                raise

            # Flag the error, but carry on
            class CustomTraceback:
                def __rich_console__(self, console: rich.console.Console, options: rich.console.ConsoleOptions):
                    sys_tb = sys.exc_info()
                    issue_url = "https://github.com/MultiQC/MultiQC/issues/new?template=bug_report.md&title={}%20module%20-%20{}".format(
                        this_module, sys_tb[0].__name__
                    )
                    yield (
                        "Please copy this log and report it at [bright_blue][link={}]https://github.com/MultiQC/MultiQC/issues[/link][/] \n"
                        "[bold underline]Please attach a file that triggers the error.[/] The last file found was: [green]{}[/]\n".format(
                            issue_url, report.last_found_file
                        )
                    )
                    yield Syntax(traceback.format_exc(), "python")

                def __rich_measure__(self, console: rich.console.Console, options: rich.console.ConsoleOptions):
                    tb_width = max([len(line) for line in traceback.format_exc().split("\n")])
                    try:
                        log_width = 71 + len(report.last_found_file)
                    except TypeError:
                        log_width = 71
                    panel_width = max(tb_width, log_width)
                    return rich.console.Measurement(panel_width, panel_width)

            console = rich.console.Console(
                stderr=True,
                force_terminal=util_functions.force_term_colors(),
                color_system=None if config.no_ansi else "auto",
            )
            console.print(
                rich.panel.Panel(
                    CustomTraceback(),
                    title=f"Oops! The '[underline]{this_module}[/]' MultiQC module broke...",
                    expand=False,
                    border_style="red",
                    style="on #272822",
                )
            )
            # Still log.debug this so that it ends up in the log file - above is just stderr for now
            logger.debug(
                f"Oops! The '{this_module}' MultiQC module broke...\n"
                + ("=" * 80)
                + "\n"
                + traceback.format_exc()
                + ("=" * 80)
            )
            # Exit code 1 for CI failures etc
            sys_exit_code = 1

        report.runtimes["mods"][run_module_names[mod_idx]] = time.time() - mod_starttime
    report.runtimes["total_mods"] = time.time() - total_mods_starttime

    # Again, if config.require_logs is set, check if for all explicitly requested
    # modules samples were found.
    if not _required_logs_found([m.anchor for m in report.modules_output]):
        return {"report": report, "config": config, "sys_exit_code": 1}

    # Update report with software versions provided in configs
    software_versions.update_versions_from_config(config, report)

    # Add section for software versions if any are found
    if not config.skip_versions_section and report.software_versions:
        # Importing here to avoid circular imports
        from multiqc.modules.software_versions import MultiqcModule

        report.modules_output.append(MultiqcModule())

    # Special-case module if we want to profile the MultiQC running time
    if config.profile_runtime:
        from multiqc.modules.profile_runtime import MultiqcModule

        report.modules_output.append(MultiqcModule())

    # Did we find anything?
    if len(report.modules_output) == 0:
        logger.warning("No analysis results found. Cleaning up..")
        shutil.rmtree(tmp_dir)
        logger.info("MultiQC complete")
        # Exit with an error code if a module broke
        return {"report": report, "config": config, "sys_exit_code": sys_exit_code}

    if config.make_report:
        # Sort the report module output if we have a config
        if len(getattr(config, "report_section_order", {})) > 0:
            section_id_order = {}
            idx = 10
            for mod in reversed(report.modules_output):
                section_id_order[mod.anchor] = idx
                idx += 10
            for anchor, ss in config.report_section_order.items():
                if anchor not in section_id_order.keys():
                    logger.debug(f"Reordering sections: anchor '{anchor}' not found.")
                    continue
                if ss.get("order") is not None:
                    section_id_order[anchor] = ss["order"]
                if ss.get("after") in section_id_order.keys():
                    section_id_order[anchor] = section_id_order[ss["after"]] + 1
                if ss.get("before") in section_id_order.keys():
                    section_id_order[anchor] = section_id_order[ss["before"]] - 1
            sorted_ids = sorted(section_id_order, key=section_id_order.get)
            report.modules_output = [
                mod for i in reversed(sorted_ids) for mod in report.modules_output if mod.anchor == i
            ]

        # Sort the report sections if we have a config
        # Basically the same as above, but sections within a module
        if len(getattr(config, "report_section_order", {})) > 0:
            # Go through each module
            for midx, mod in enumerate(report.modules_output):
                section_id_order = {}
                # Get a list of the section anchors
                idx = 10
                for s in mod.sections:
                    section_id_order[s["anchor"]] = idx
                    idx += 10
                # Go through each section to be reordered
                for anchor, ss in config.report_section_order.items():
                    # Section to be moved is not in this module
                    if anchor not in section_id_order.keys():
                        logger.debug(f"Reordering sections: anchor '{anchor}' not found for module '{mod.name}'.")
                        continue
                    if ss == "remove":
                        section_id_order[anchor] = False
                        continue
                    if ss.get("order") is not None:
                        section_id_order[anchor] = ss["order"]
                    if ss.get("after") in section_id_order.keys():
                        section_id_order[anchor] = section_id_order[ss["after"]] + 1
                    if ss.get("before") in section_id_order.keys():
                        section_id_order[anchor] = section_id_order[ss["before"]] - 1
                # Remove module sections
                section_id_order = {s: o for s, o in section_id_order.items() if o is not False}
                # Sort the module sections
                sorted_ids = sorted(section_id_order, key=section_id_order.get)
                report.modules_output[midx].sections = [s for i in sorted_ids for s in mod.sections if s["anchor"] == i]

    plugin_hooks.mqc_trigger("after_modules")

    # Remove empty data sections from the General Stats table
    empty_keys = [i for i, d in enumerate(report.general_stats_data[:]) if len(d) == 0]
    empty_keys.sort(reverse=True)
    for i in empty_keys:
        del report.general_stats_data[i]
        del report.general_stats_headers[i]

    # Add general-stats IDs to table row headers
    for idx, h in enumerate(report.general_stats_headers):
        for k in h.keys():
            if "rid" not in h[k]:
                h[k]["rid"] = re.sub(r"\W+", "_", k).strip().strip("_")
            ns_html = re.sub(r"\W+", "_", h[k]["namespace"]).strip().strip("_").lower()
            report.general_stats_headers[idx][k]["rid"] = report.save_htmlid(
                f"mqc-generalstats-{ns_html}-{h[k]['rid']}"
            )

    all_hidden = True
    for headers in report.general_stats_headers:
        for h in headers.values():
            if not h.get("hidden", False):
                all_hidden = False
                break

    # Generate the General Statistics HTML & write to file
    if len(report.general_stats_data) > 0 and not config.skip_generalstats and not all_hidden:
        pconfig = {
            "id": "general_stats_table",
            "table_title": "General Statistics",
            "save_file": True,
            "raw_data_fn": "multiqc_general_stats",
        }
        report.general_stats_html = table.plot(report.general_stats_data, report.general_stats_headers, pconfig)
    else:
        config.skip_generalstats = True

    if config.data_dir is not None:
        # Write the report sources to disk
        report.data_sources_tofile()

        # Create a file with the module DOIs
        report.dois_tofile(report.modules_output)

    if config.make_report:
        # Compress the report plot JSON data
        runtime_compression_start = time.time()
        logger.debug("Compressing plot data")
        report.plot_compressed_json = report.compress_json(report.plot_data)
        report.runtimes["total_compression"] = time.time() - runtime_compression_start

    plugin_hooks.mqc_trigger("before_report_generation")

    # Data Export / MegaQC integration - save report data to file or send report data to an API endpoint
    if config.data_dump_file or (config.megaqc_url and config.megaqc_upload):
        multiqc_json_dump = util_functions.multiqc_dump_json(report)
        if config.data_dump_file:
            util_functions.write_data_file(multiqc_json_dump, "multiqc_data", False, "json")
        if config.megaqc_url:
            megaqc.multiqc_api_post(multiqc_json_dump)

    if config.development:
        with open(os.path.join(config.data_dir, "multiqc_plots.js"), "w") as f:
            f.write(json.dumps(report.plot_data))

    # Make the final report path & data directories
    if filename != "stdout":
        if config.make_report:
            config.output_fn = os.path.join(config.output_dir, config.output_fn_name)
        config.data_dir = os.path.join(config.output_dir, config.data_dir_name)
        config.plots_dir = os.path.join(config.output_dir, config.plots_dir_name)
        # del config.data_dir_name
        # del config.plots_dir_name
        deleted_report = False
        deleted_data_dir = False
        deleted_export_plots = False
        # Check for existing reports and remove if -f was specified
        if (
            (config.make_report and os.path.exists(config.output_fn))
            or (config.make_data_dir and os.path.exists(config.data_dir))
            or (config.export_plots and os.path.exists(config.plots_dir))
        ):
            if config.force:
                if config.make_report and os.path.exists(config.output_fn):
                    deleted_report = True
                    os.remove(config.output_fn)
                if config.make_data_dir and os.path.exists(config.data_dir):
                    deleted_data_dir = True
                    shutil.rmtree(config.data_dir)
                if config.export_plots and os.path.exists(config.plots_dir):
                    deleted_export_plots = True
                    shutil.rmtree(config.plots_dir)
            else:
                # Set up the base names of the report and the data dir
                report_num = 1
                dir_base = os.path.basename(config.data_dir)
                plots_base = os.path.basename(config.plots_dir)

                # Iterate through appended numbers until we find one that's free
                while (
                    (config.make_report and os.path.exists(config.output_fn))
                    or (config.make_data_dir and os.path.exists(config.data_dir))
                    or (config.export_plots and os.path.exists(config.plots_dir))
                ):
                    if config.make_report:
                        report_base, report_ext = os.path.splitext(config.output_fn_name)
                        config.output_fn = os.path.join(config.output_dir, f"{report_base}_{report_num}{report_ext}")
                    config.data_dir = os.path.join(config.output_dir, f"{dir_base}_{report_num}")
                    config.plots_dir = os.path.join(config.output_dir, f"{plots_base}_{report_num}")
                    report_num += 1
                if config.make_report:
                    config.output_fn_name = os.path.basename(config.output_fn)
                config.data_dir_name = os.path.basename(config.data_dir)
                config.plots_dir_name = os.path.basename(config.plots_dir)
                logger.info("Existing reports found, adding suffix to filenames. Use '--force' to overwrite.")

        # Make directories for report if needed
        if config.make_report:
            if not os.path.exists(os.path.dirname(config.output_fn)):
                os.makedirs(os.path.dirname(config.output_fn))
            logger.info(
                "Report      : {}{}".format(
                    os.path.relpath(config.output_fn),
                    "   (overwritten)" if deleted_report else "",
                )
            )
        else:
            logger.info("Report      : None")

        if config.make_data_dir is False:
            logger.info("Data        : None")
        else:
            # Make directories for data_dir
            logger.info(
                "Data        : {}{}".format(
                    os.path.relpath(config.data_dir),
                    "   (overwritten)" if deleted_data_dir else "",
                )
            )
            # Modules have run, so data directory should be complete by now. Move its contents.
            logger.debug(f"Moving data file from '{config.data_tmp_dir}' to '{config.data_dir}'")
            shutil.copytree(
                config.data_tmp_dir,
                config.data_dir,
                # Override default shutil.copy2 function to copy files. The default
                # function copies times and mode, which we want to avoid on purpose
                # to get around the problem with mounted CIFS shares (see #625).
                # shutil.copyfile only copies the file without any metadata.
                copy_function=shutil.copyfile,
            )
            shutil.rmtree(config.data_tmp_dir)

        if config.output_fn is not None:
            logger.debug(f"Full report path: {os.path.realpath(config.output_fn)}")

        # Copy across the static plot images if requested
        if config.export_plots:
            config.plots_dir = os.path.join(config.output_dir, config.plots_dir_name)
            if os.path.exists(config.plots_dir):
                if config.force:
                    shutil.rmtree(config.plots_dir)
                else:
                    logger.error(f"Output directory {config.plots_dir} already exists.")
                    logger.info("Use -f or --force to overwrite existing reports")
                    shutil.rmtree(tmp_dir)
                    return {"report": report, "config": config, "sys_exit_code": 1}
            logger.info(
                "Plots       : {}{}".format(
                    os.path.relpath(config.plots_dir),
                    "   (overwritten)" if deleted_export_plots else "",
                )
            )

            # Modules have run, so plots directory should be complete by now. Move its contents.
            logger.debug(f"Moving plots directory from '{config.plots_tmp_dir}' to '{config.plots_dir}'")
            shutil.copytree(
                config.plots_tmp_dir,
                config.plots_dir,
                # Override default shutil.copy2 function to copy files. The default
                # function copies times and mode, which we want to avoid on purpose
                # to get around the problem with mounted CIFS shares (see #625).
                # shutil.copyfile only copies the file without any metadata.
                copy_function=shutil.copyfile,
            )
            shutil.rmtree(config.plots_tmp_dir)

    plugin_hooks.mqc_trigger("before_template")

    # Generate report if required
    if config.make_report:
        # Load in parent template files first if a child theme
        parent_template = None
        try:
            parent_template = config.avail_templates[template_mod.template_parent].load()
        except AttributeError:
            pass  # Not a child theme
        else:
            shutil.copytree(parent_template.template_dir, tmp_dir, dirs_exist_ok=True)

        # Copy the template files to the tmp directory (`dirs_exist_ok` makes sure
        # parent template files are overwritten)
        shutil.copytree(template_mod.template_dir, tmp_dir, dirs_exist_ok=True)

        # Function to include file contents in Jinja template
        def include_file(name, fdir=tmp_dir, b64=False):
            try:
                if fdir is None:
                    fdir = ""
                path = os.path.join(fdir, name)

                if config.development:
                    if os.path.exists(dev_path := os.path.join(template_mod.template_dir, name)):
                        fdir = template_mod.template_dir
                        name = dev_path
                        path = dev_path
                    elif parent_template and os.path.exists(
                        dev_path := os.path.join(parent_template.template_dir, name)
                    ):
                        fdir = template_mod.template_dir
                        name = dev_path
                        path = dev_path

                    if re.match(r".*\.min\.(js|css)$", name):
                        unminimized_name = re.sub(r"\.min\.", ".", name)
                        if os.path.exists(os.path.join(fdir, unminimized_name)):
                            name = unminimized_name

                    if name.endswith(".js"):
                        return f'</script><script type="text/javascript" src="{name}">'
                    if name.endswith(".css"):
                        return f'</style><link rel="stylesheet" href="{name}">'

                if b64:
                    with io.open(path, "rb") as f:
                        return base64.b64encode(f.read()).decode("utf-8")
                else:
                    with io.open(path, "r", encoding="utf-8") as f:
                        return f.read()
            except (OSError, IOError) as e:
                logger.error(f"Could not include file '{name}': {e}")

        # Load the report template
        try:
            env = jinja2.Environment(loader=jinja2.FileSystemLoader(tmp_dir))
            env.globals["include_file"] = include_file
            j_template = env.get_template(template_mod.base_fn, globals={"development": config.development})
        except:  # noqa: E722
            raise IOError(f"Could not load {config.template} template file '{template_mod.base_fn}'")

        # Use jinja2 to render the template and overwrite
        config.analysis_dir = [os.path.realpath(d) for d in config.analysis_dir]
        report_output = j_template.render(report=report, config=config)
        if filename == "stdout":
            print(report_output.encode("utf-8"), file=sys.stdout)
        else:
            try:
                with io.open(config.output_fn, "w", encoding="utf-8") as f:
                    print(report_output, file=f)
            except IOError as e:
                raise IOError(f"Could not print report to '{config.output_fn}' - {IOError(e)}")

            # Copy over files if requested by the theme
            try:
                for f in template_mod.copy_files:
                    fn = os.path.join(tmp_dir, f)
                    dest_dir = os.path.join(os.path.dirname(config.output_fn), f)
                    shutil.copytree(fn, dest_dir, dirs_exist_ok=True)
            except AttributeError:
                pass  # No files to copy

    # Clean up temporary directory
    shutil.rmtree(tmp_dir)

    # Zip the data directory if requested
    if config.zip_data_dir and config.data_dir is not None:
        shutil.make_archive(config.data_dir, "zip", config.data_dir)
        shutil.rmtree(config.data_dir)

    # Try to create a PDF if requested
    if make_pdf:
        try:
            pdf_fn_name = config.output_fn.replace(".html", ".pdf")
            pandoc_call = [
                "pandoc",
                "--standalone",
                config.output_fn,
                "--output",
                pdf_fn_name,
                "--pdf-engine=xelatex",
                "-V",
                "documentclass=article",
                "-V",
                "geometry=margin=1in",
                "-V",
                "title=",
            ]
            if config.pandoc_template is not None:
                pandoc_call.append(f"--template={config.pandoc_template}")
            logger.debug(f"Attempting Pandoc conversion to PDF with following command:\n{' '.join(pandoc_call)}")
            pdf_exit_code = subprocess.call(pandoc_call)
            if pdf_exit_code != 0:
                logger.error("Error creating PDF! Pandoc returned a non-zero exit code.")
            else:
                logger.info(f"PDF Report  : {pdf_fn_name}")
        except OSError as e:
            if e.errno == errno.ENOENT:
                logger.error("Error creating PDF - pandoc not found. Is it installed? http://pandoc.org/")
            else:
                logger.error(
                    "Error creating PDF! Something went wrong when creating the PDF\n"
                    + ("=" * 60)
                    + f"\n{traceback.format_exc()}\n"
                    + ("=" * 60)
                )

    plugin_hooks.mqc_trigger("execution_finish")

    report.runtimes["total"] = time.time() - start_execution_time
    if config.profile_runtime:
        logger.warning(f"Run took {report.runtimes['total']:.2f} seconds")
        logger.warning(f" - {report.runtimes['total_sp']:.2f}s: Searching files")
        logger.warning(f" - {report.runtimes['total_mods']:.2f}s: Running modules")
        if config.make_report:
            logger.warning(f" - {report.runtimes['total_compression']:.2f}s: Compressing report data")
            logger.info(f"For more information, see the 'Run Time' section in {os.path.relpath(config.output_fn)}")

    if report.num_mpl_plots > 0 and not config.plots_force_flat:
        if not config.plots_force_interactive:
            console.print(
                "[blue]|           multiqc[/] | "
                "Flat-image plots used. Disable with '--interactive'. "
                "See [link=https://multiqc.info/docs/#flat--interactive-plots]docs[/link]."
            )

    if config.strict and len(report.lint_errors) > 0:
        logger.error(f"Found {len(report.lint_errors)} linting errors!\n" + "\n".join(report.lint_errors))
        sys_exit_code = 1

    logger.info("MultiQC complete")

    # Move the log file into the data directory
    log.move_tmp_log(logger)

    # Return the running information from the run:
    #
    # * report instance
    # * config instance
    # * appropriate error code (eg. 1 if a module broke, 0 on success)
    #
    return {"report": report, "config": config, "sys_exit_code": sys_exit_code}


def _required_logs_found(modules_with_logs):
    if config.require_logs:
        required_modules_with_no_logs = [
            m
            for m in getattr(config, "run_modules", [])
            if m.lower() not in [m.lower() for m in modules_with_logs]
            and m.lower() not in getattr(config, "exclude_modules", [])
        ]
        if required_modules_with_no_logs:
            logger.critical(
                "The following modules were explicitly requested but no log files were found: {}".format(
                    ", ".join(required_modules_with_no_logs)
                )
            )
            return False
    return True



""" MultiQC module to parse output from pycoQC """

import logging

import yaml

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph, table

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="pycoQC",
            anchor="pycoqc",
            href="https://github.com/tleonardi/pycoQC",
            info="computes metrics and generates interactive QC plots for Oxford Nanopore technologies sequencing data",
            doi="10.21105/joss.01236",
        )

        self.pycoqc_data = {}
        for f in self.find_log_files("pycoqc"):
            if f["s_name"] in self.pycoqc_data:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {f['s_name']}")
            data = self.load_data(f["f"])
            # Function can return None if YAML parsing failed
            if data:
                self.pycoqc_data[f["s_name"]] = data
                self.add_software_version(data["pycoqc"]["version"], f["s_name"])
            self.add_data_source(f)

        self.pycoqc_data = self.ignore_samples(self.pycoqc_data)

        # Stop if we didn't find anything
        if len(self.pycoqc_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.pycoqc_data)} reports")

        # Write data to file
        self.write_data_file(self.pycoqc_data, "pycoqc")

        self.parse_data()

        self.make_general_stats()
        self.make_pycoqc_table()
        self.read_base_count_plot()

        # Only add these sections if we have data (not available for older versions of PycoQC)
        if self.read_length_plot_data[0]:
            self.read_length_plot()

        if self.quality_plot_data[0]:
            self.make_quality_plot()

    def load_data(self, f):
        """Load the PycoQC YAML file"""
        try:
            return yaml.load(f, Loader=yaml.SafeLoader)
        except Exception as e:
            log.warning(f"Could not parse YAML for '{f}': \n  {e}")
            return None

    def parse_data(self):
        """Convert the parsed data into the correct structures for MultiQC functions"""

        self.table_data = {}
        self.reads_data = {}
        self.bases_data = {}
        length_plot_all = {}
        length_plot_pass = {}
        qual_plot_all = {}
        qual_plot_pass = {}

        for sample, sample_data in self.pycoqc_data.items():
            self.table_data[sample] = {
                "all_median_read_length": sample_data["All Reads"]["basecall"]["len_percentiles"][50],
                "all_median_phred_score": sample_data["All Reads"]["basecall"]["qual_score_percentiles"][50],
                "all_n50": sample_data["All Reads"]["basecall"]["N50"],
                "all_run_duration": sample_data["All Reads"]["run"]["run_duration"],
                "all_channels": sample_data["All Reads"]["run"]["active_channels"],
                "all_reads": sample_data["All Reads"]["basecall"]["reads_number"],
                "all_bases": sample_data["All Reads"]["basecall"]["bases_number"],
                "passed_median_read_length": sample_data["Pass Reads"]["basecall"]["len_percentiles"][50],
                "passed_median_phred_score": sample_data["Pass Reads"]["basecall"]["qual_score_percentiles"][50],
                "passed_n50": sample_data["Pass Reads"]["basecall"]["N50"],
                "passed_channels": sample_data["Pass Reads"]["run"]["active_channels"],
                "passed_reads": sample_data["Pass Reads"]["basecall"]["reads_number"],
                "passed_bases": sample_data["Pass Reads"]["basecall"]["bases_number"],
            }

            self.reads_data[sample] = {
                "passed_reads": sample_data["Pass Reads"]["basecall"]["reads_number"],
                "non_passed_reads": sample_data["All Reads"]["basecall"]["reads_number"]
                - sample_data["Pass Reads"]["basecall"]["reads_number"],
            }

            self.bases_data[sample] = {
                "passed_bases": sample_data["Pass Reads"]["basecall"]["bases_number"],
                "non_passed_bases": sample_data["All Reads"]["basecall"]["bases_number"]
                - sample_data["Pass Reads"]["basecall"]["bases_number"],
            }
            try:
                length_x_vals_all = sample_data["All Reads"]["basecall"]["len_hist"]["x"]
                length_y_vals_all = sample_data["All Reads"]["basecall"]["len_hist"]["y"]
                length_plot_all[sample] = dict(zip(length_x_vals_all, length_y_vals_all))
                length_x_vals_pass = sample_data["Pass Reads"]["basecall"]["len_hist"]["x"]
                length_y_vals_pass = sample_data["Pass Reads"]["basecall"]["len_hist"]["y"]
                length_plot_pass[sample] = dict(zip(length_x_vals_pass, length_y_vals_pass))

                qual_x_vals_all = sample_data["All Reads"]["basecall"]["qual_score_hist"]["x"]
                qual_y_vals_all = sample_data["All Reads"]["basecall"]["qual_score_hist"]["y"]
                qual_plot_all[sample] = dict(zip(qual_x_vals_all, qual_y_vals_all))
                qual_x_vals_pass = sample_data["Pass Reads"]["basecall"]["qual_score_hist"]["x"]
                qual_y_vals_pass = sample_data["Pass Reads"]["basecall"]["qual_score_hist"]["y"]
                qual_plot_pass[sample] = dict(zip(qual_x_vals_pass, qual_y_vals_pass))
            except KeyError:
                log.debug(
                    "'{}': Could not find plot data. Please make sure you are using pycoQC v2.5.0.20 or newer.".format(
                        sample
                    )
                )

        self.read_length_plot_data = [length_plot_pass, length_plot_all]
        self.quality_plot_data = [qual_plot_pass, qual_plot_all]

    def make_general_stats(self):
        general_stats_headers = {
            "passed_median_read_length": {
                "title": "Read Length - Pass (bp)",
                "description": "Median Read Length - passing reads (base pairs)",
                "scale": "BuPu",
                "shared_key": "median_read_len",
                "format": "{:,.0f}",
            },
            "all_median_read_length": {
                "title": "Read Length - All (bp)",
                "description": "Median Read Length - all reads (base pairs)",
                "scale": "BuPu",
                "shared_key": "median_read_len",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "passed_reads": {
                "title": f"{config.long_read_count_prefix} Reads - Pass",
                "description": f"Number of reads - passing reads ({config.long_read_count_desc})",
                "scale": "BuGn",
                "modify": lambda x: x * config.long_read_count_multiplier,
                "shared_key": "long_read_count",
            },
            "all_reads": {
                "title": f"{config.long_read_count_prefix} Reads - All",
                "description": f"Number of reads - all reads ({config.long_read_count_desc})",
                "scale": "BuGn",
                "modify": lambda x: x * config.long_read_count_multiplier,
                "shared_key": "long_read_count",
                "hidden": True,
            },
            "passed_bases": {
                "title": f"{config.base_count_prefix} Bases - Pass",
                "description": f"Number of bases - passing reads ({config.base_count_desc} of base pairs)",
                "scale": "OrRd",
                "modify": lambda x: x * config.base_count_multiplier,
                "shared_key": "base_count",
            },
            "all_bases": {
                "title": f"{config.base_count_prefix} Bases - All",
                "description": f"Number of bases - all reads ({config.base_count_desc} of base pairs)",
                "scale": "OrRd",
                "modify": lambda x: x * config.base_count_multiplier,
                "shared_key": "base_count",
                "hidden": True,
            },
        }

        self.general_stats_addcols(self.table_data, general_stats_headers)

    def make_pycoqc_table(self):
        pycoqc_table_headers = {
            "passed_n50": {
                "title": "N50 - Pass (bp)",
                "description": "N50 - passing reads (base pairs)",
                "scale": "BuGn",
                "shared_key": "n50",
                "format": "{:,.0f}",
            },
            "all_n50": {
                "title": "N50 - All (bp)",
                "description": "N50 - all reads (base pairs)",
                "scale": "BuGn",
                "shared_key": "n50",
                "format": "{:,.0f}",
            },
            "passed_median_phred_score": {
                "title": "Median read qual - Pass",
                "description": "Median PHRED quality score - passing reads",
                "scale": "PuRd",
                "shared_key": "phred",
            },
            "all_median_phred_score": {
                "title": "Median read qual - All",
                "description": "Median PHRED quality score - all reads",
                "scale": "PuRd",
                "shared_key": "phred",
            },
            "passed_channels": {
                "title": "Active Channels - Pass",
                "description": "Number of active channels - passing reads",
                "scale": "OrRd",
                "shared_key": "channels",
                "format": "{:,.0f}",
            },
            "all_channels": {
                "title": "Active Channels - All",
                "description": "Number of active channels - all reads",
                "scale": "OrRd",
                "shared_key": "channels",
                "format": "{:,.0f}",
            },
            "all_run_duration": {
                "title": "Run duration (h)",
                "description": "Run duration (hours)",
                "scale": "PuBuGn",
            },
        }

        pycoqc_table_config = {
            "id": "pycoqc_stats_table",
            "namespace": "pycoQC",
            "title": "pycoQC: Statistics",
        }

        self.add_section(
            name="Statistics",
            anchor="pycoqc_stats",
            plot=table.plot(self.table_data, pycoqc_table_headers, pycoqc_table_config),
        )

    def read_base_count_plot(self):
        """Make a section with a bar plot showing the counts of reads and bases
        before and after filtering"""

        # Plot config
        read_bar_config = {
            "id": "pycoqc_count_plot",
            "title": "pycoQC: Read Counts",
            "ylab": "Sample",
            "data_labels": [
                {"name": "Reads", "ylab": "Number of Reads"},
                {"name": "Bases", "ylab": "Number of Base Pairs"},
            ],
        }

        # Two sets of categories for the two datasets
        read_cats = {
            "passed_reads": {"name": "Passed Reads"},
            "non_passed_reads": {"name": "Failed Reads"},
        }
        base_cats = {
            "passed_bases": {"name": "Passed Bases"},
            "non_passed_bases": {"name": "Failed Bases"},
        }

        # Make the plot
        bargraph_plot = bargraph.plot([self.reads_data, self.bases_data], [read_cats, base_cats], read_bar_config)

        self.add_section(
            name="Read / Base counts",
            anchor="pycoqc_counts",
            description="Number of sequenced reads / bases passing and failing QC thresholds.",
            plot=bargraph_plot,
        )

    def read_length_plot(self):
        """Read length plot, showing passed reads and all reads"""

        read_length_config = {
            "id": "pycoqc_read_len_plot",
            "title": "pycoQC: Read Length",
            "ylab": "Read Density",
            "xlab": "Basecalled Length (bp)",
            "xLog": True,
            "data_labels": [
                {"name": "Passing Reads", "ylab": "Read Density", "xlab": "Basecalled Length (bp)"},
                {"name": "All Reads", "ylab": "Read Density", "xlab": "Basecalled Length (bp)"},
            ],
        }
        self.add_section(
            name="Read length",
            anchor="pycoqc_read_len",
            description="Distribution of read length for all / passed reads.",
            plot=linegraph.plot(self.read_length_plot_data, read_length_config),
        )

    def make_quality_plot(self):
        """Quality plot, showing distrubtion of PHRED scores."""

        qual_config = {
            "id": "pycoqc_read_qual_plot",
            "title": "pycoQC: Read Quality Scores",
            "ylab": "Read Density",
            "xlab": "Read Quality PHRED Score",
            "xmin": 0,
            "data_labels": [
                {"name": "Passing Reads", "ylab": "Read Density", "xlab": "Read Quality PHRED Score"},
                {"name": "All Reads", "ylab": "Read Density", "xlab": "Read Quality PHRED Score"},
            ],
        }
        self.add_section(
            name="Quality scores",
            anchor="pycoqc_read_qual",
            description="Distribution of quality scores for all / passed reads.",
            plot=linegraph.plot(self.quality_plot_data, qual_config),
        )



""" MultiQC module to parse output from Kaiju """

import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Kaiju",
            anchor="kaiju",
            href="http://kaiju.binf.ku.dk/",
            info="a fast and sensitive taxonomic classification for metagenomics.",
            doi="10.1038/ncomms11257",
        )

        # Set up data structures
        self.kaiju_data = {}
        self.taxonomic_ranks = ["species", "genus", "family", "order", "class", "phylum"]

        # Find and parse kaiju2table report
        for f in self.find_log_files("kaiju", filehandles=True):
            taxo_rank, parsed_data = self.parse_kaiju2table_report(f)
            if taxo_rank is not None and parsed_data is not None:
                # One report can have several samples, raise error if sample already viewed with this taxonomic rank
                s_names = parsed_data.keys()
                for s_name in s_names:
                    if taxo_rank in self.kaiju_data and s_name in self.kaiju_data[taxo_rank].keys():
                        log.debug(
                            "Duplicate sample found in logs at {} rank! Overwriting sample: {}".format(
                                taxo_rank, s_name
                            )
                        )

                    # Superfluous function call to confirm that it is used in this module
                    # Replace None with actual version if it is available
                    self.add_software_version(None, s_name)

                self.add_data_source(f)
                if taxo_rank in self.kaiju_data.keys():
                    self.kaiju_data[taxo_rank].update(parsed_data)
                else:
                    self.kaiju_data[taxo_rank] = parsed_data

        # Filters to strip out ignored sample names
        all_samples_name = []
        for taxo_rank in self.kaiju_data.keys():
            self.kaiju_data[taxo_rank] = self.ignore_samples(self.kaiju_data[taxo_rank])
            all_samples_name.extend(self.kaiju_data[taxo_rank].keys())
        self.all_s_name = set(all_samples_name)
        # Number of samples found
        try:
            num_samples = max([len(taxo_rank) for taxo_rank in self.kaiju_data.values()])
        except ValueError:
            # No log files so didn't get any taxo_ranks
            raise ModuleNoSamplesFound

        # no file found
        if num_samples == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {num_samples} reports")

        self.kaiju_total_pct = dict()
        self.kaiju_sample_total_readcounts = dict()
        self.kaiju_sample_unclassified = dict()

        self.sum_sample_counts()
        self.kaiju_stats_table()
        self.top_five_barplot()

    def parse_kaiju2table_report(self, f):
        """Search a kaiju with a set of regexes"""
        parsed_data = {}
        taxo_rank = None

        for line in f["f"]:
            if line.startswith("file\t"):
                continue
            (s_file, pct, reads, taxon_id, taxon_names) = line.rstrip().split("\t")
            s_name = self.clean_s_name(s_file, f)
            if s_name not in parsed_data.keys():
                parsed_data[s_name] = {"assigned": {}}

            if taxon_names.startswith("cannot be assigned") or taxon_names == "unclassified":
                if taxon_names.startswith("cannot be assigned"):
                    parsed_data[s_name]["cannot be assigned"] = {"count": int(reads), "percent": float(pct)}
                    taxo_rank = taxon_names.split()[-1]
                else:
                    parsed_data[s_name]["unclassified"] = {"count": int(reads), "percent": float(pct)}
            else:
                parsed_data[s_name]["assigned"][taxon_names] = {"count": int(reads), "percent": float(pct)}

        return taxo_rank, parsed_data

    def sum_sample_counts(self):
        """Sum counts across all samples for kaiju data"""

        # Sum the percentages for each taxa across all samples
        # Allows us to pick top-5 for each rank
        # Use percentages instead of counts so that deeply-sequences samples
        # are not unfairly over-represented
        first_pass_total = False

        for rank_name, data in self.kaiju_data.items():
            for s_name, samples_values in data.items():
                # perform sum at first level only
                if rank_name not in self.kaiju_total_pct:
                    self.kaiju_total_pct[rank_name] = dict()
                # If total did not already compute
                if s_name not in self.kaiju_sample_total_readcounts:
                    first_pass_total = True
                    self.kaiju_sample_total_readcounts[s_name] = 0

                if "assigned" in samples_values:
                    for classif, row in samples_values["assigned"].items():
                        if classif not in self.kaiju_total_pct[rank_name]:
                            self.kaiju_total_pct[rank_name][classif] = 0
                        self.kaiju_total_pct[rank_name][classif] += row["percent"]
                        if first_pass_total:
                            self.kaiju_sample_total_readcounts[s_name] += row["count"]
                    if first_pass_total:
                        if "cannot be assigned" in samples_values:
                            self.kaiju_sample_total_readcounts[s_name] += samples_values["cannot be assigned"]["count"]
                        if "unclassified" in samples_values:
                            self.kaiju_sample_total_readcounts[s_name] += samples_values["unclassified"]["count"]
                            self.kaiju_sample_unclassified[s_name] = samples_values["unclassified"]["count"]
                        first_pass_total = False

    def kaiju_stats_table(self):
        """Take the parsed stats from the Kaiju reports and add them to the
        basic stats table at the top of the report"""
        headers = {}
        taxo_ranks = sorted(self.kaiju_data)
        # print only assigned at phylum rank in general table or the first available.
        if len(taxo_ranks) >= 1 and "phylum" in taxo_ranks:
            general_taxo_rank = "Phylum"
        else:
            general_taxo_rank = taxo_ranks[0].capitalize()

        headers["% Assigned"] = {
            "title": "% Reads assigned",
            "description": f"Percentage of reads assigned at {general_taxo_rank} rank",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "scale": "RdYlGn",
        }
        headers["assigned"] = {
            "title": f"{config.read_count_prefix} Reads assigned",
            "description": "Number of reads assigned ({})  at {} rank".format(
                config.read_count_desc, general_taxo_rank
            ),
            "modify": lambda x: x * config.read_count_multiplier,
            "scale": "Blues",
        }
        headers["% Unclassified"] = {
            "title": "% Unclassified",
            "description": "Percentage of reads that were unclassified",
            "suffix": "%",
            "max": 100,
            "scale": "OrRd",
        }
        tdata = {}
        for s_name, d in self.kaiju_sample_total_readcounts.items():
            tdata[s_name] = {
                # Default values in case if data is not available for this sample at this rank
                "assigned": 0,
                "% Assigned": 0,
            }
            total = self.kaiju_sample_total_readcounts[s_name]
            if total == 0:
                continue
            tdata[s_name]["% Unclassified"] = self.kaiju_sample_unclassified[s_name] * 100 / total
            if s_name in self.kaiju_data[general_taxo_rank.lower()]:
                tdata[s_name]["assigned"] = (
                    total
                    - self.kaiju_sample_unclassified[s_name]
                    - self.kaiju_data[general_taxo_rank.lower()][s_name]["cannot be assigned"]["count"]
                )
                tdata[s_name]["% Assigned"] = tdata[s_name]["assigned"] * 100 / total
        self.general_stats_addcols(tdata, headers)

    def top_five_barplot(self):
        """Add a bar plot showing the top-5 from each taxa rank"""

        # ordered rank used
        rank_used = []
        for rank in self.taxonomic_ranks:
            if rank in self.kaiju_data:
                rank_used.append(rank)

        pconfig = {
            "id": "kaiju-topfive-plot",
            "title": "Kaiju: Top taxa",
            "ylab": "Number of reads",
            "cpswitch_counts_label": "Number of reads",
            "data_labels": [
                {"name": rank.capitalize(), "ylab": f"Number of Reads at {rank.capitalize()} rank"}
                for rank in rank_used
            ],
        }

        pd = []
        cats = []

        for rank_name in rank_used:
            rank_cats = {}
            rank_data = {}
            # Loop through the summed tax percentages to get the top 5 across all samples

            top_five_sorted_pct = sorted(self.kaiju_total_pct[rank_name].items(), key=lambda x: x[1], reverse=True)[0:5]

            # retrieve top five counts
            counts_shown = {}
            for classif, pct_sum in top_five_sorted_pct:
                rank_cats[classif] = {"name": classif}
                # Pull out counts for this rank + classif from each sample
                for s_name, d in self.kaiju_data[rank_name].items():
                    if s_name not in rank_data:
                        rank_data[s_name] = dict()
                    if s_name not in counts_shown:
                        counts_shown[s_name] = 0
                    if classif in d["assigned"]:
                        if classif not in rank_data[s_name]:
                            rank_data[s_name][classif] = 0
                        rank_data[s_name][classif] += d["assigned"][classif]["count"]
                        counts_shown[s_name] += d["assigned"][classif]["count"]

            # Add in unclassified/cannot assigned/other
            for s_name in self.all_s_name:
                if s_name in self.kaiju_data[rank_name]:
                    d = self.kaiju_data[rank_name][s_name]
                    rank_data[s_name]["not assigned"] = d["cannot be assigned"]["count"]
                    rank_data[s_name]["unclassified"] = d["unclassified"]["count"]
                    rank_data[s_name]["other"] = (
                        self.kaiju_sample_total_readcounts[s_name]
                        - counts_shown[s_name]
                        - d["cannot be assigned"]["count"]
                        - d["unclassified"]["count"]
                    )
                    # This should never happen... But it does sometimes if the total read count is a bit off
                    if rank_data[s_name]["other"] < 0:
                        log.debug(
                            "Found negative 'other' count for {} ({}): {}".format(
                                s_name, rank_name, rank_data[s_name]["other"]
                            )
                        )
                        rank_data[s_name]["other"] = 0

                else:
                    rank_data[s_name] = {}
                    rank_data[s_name]["other"] = 0
                    rank_data[s_name]["missing info"] = (
                        self.kaiju_sample_total_readcounts[s_name] - self.kaiju_sample_unclassified[s_name]
                    )
                    rank_data[s_name]["unclassified"] = self.kaiju_sample_unclassified[s_name]

            rank_cats["other"] = {"name": "Other", "color": "#a65628"}
            rank_cats["not assigned"] = {"name": "Cannot be assigned", "color": "#cccccc"}
            rank_cats["missing info"] = {"name": "Missing info", "color": "#979a9a"}
            rank_cats["unclassified"] = {"name": "Unclassified", "color": "#d4949c"}

            self.write_data_file(rank_data, "multiqc_kaiju_" + rank_name)
            cats.append(rank_cats)
            pd.append(rank_data)

        if len(cats) == 0:
            log.debug("No data for Kaiju top five barplot. Skipping.")
            return

        self.add_section(
            name="Top taxa",
            anchor="kaiju-topfive",
            description="The number of reads falling into the top 5 taxa across different ranks.",
            helptext="""
                To make this plot, the percentage of each sample assigned to a given taxa is summed across all samples.
                The counts for these top five taxa are then plotted for each of the taxa ranks found in logs.
                The unclassified count is always shown across all taxa ranks.
                The 'Cannot be assigned' count correspond to reads classified but not at this taxa rank.

                The category _"Other"_ shows the difference between the above total assingned read count and the sum of the read counts
                in the top 5 taxa shown. This should cover all taxa _not_ in the top 5.
            """,
            plot=bargraph.plot(pd, cats, pconfig),
        )



""" MultiQC module to parse output from Samtools """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

from .flagstat import FlagstatReportMixin
from .idxstats import IdxstatsReportMixin
from .rmdup import RmdupReportMixin
from .coverage import CoverageReportMixin

# Import the Samtools submodules
from .stats import StatsReportMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(
    BaseMultiqcModule, StatsReportMixin, FlagstatReportMixin, IdxstatsReportMixin, RmdupReportMixin, CoverageReportMixin
):
    """Samtools has a number of different commands and outputs.
    This MultiQC module supports some but not all. The code for
    each script is split into its own file and adds a section to
    the module output if logs are found."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Samtools",
            anchor="Samtools",
            target="Samtools",
            href="http://www.htslib.org",
            info=" is a suite of programs for interacting with high-throughput sequencing data.",
            doi="10.1093/bioinformatics/btp352",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # Call submodule functions
        n["stats"] = self.parse_samtools_stats()
        if n["stats"] > 0:
            log.info(f"Found {n['stats']} stats reports")

        n["flagstat"] = self.parse_samtools_flagstats()
        if n["flagstat"] > 0:
            log.info(f"Found {n['flagstat']} flagstat reports")

        n["idxstats"] = self.parse_samtools_idxstats()
        if n["idxstats"] > 0:
            log.info(f"Found {n['idxstats']} idxstats reports")

        n["rmdup"] = self.parse_samtools_rmdup()
        if n["rmdup"] > 0:
            log.info(f"Found {n['rmdup']} rmdup reports")

        n["coverage"] = self.parse_samtools_coverage()
        if n["coverage"] > 0:
            log.info(f"Found {n['coverage']} coverage reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)
